# Copyright 2019 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy of the
# License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from ctypes import byref, cdll, c_int, POINTER, c_float, c_char_p, c_double, create_string_buffer, c_int32, c_void_p
import numpy as np
from numpy.ctypeslib import ndpointer
import platform, ctypes

# Some useful constants
deg2rad = np.deg2rad(1)
rad2deg = np.rad2deg(1)
rpm2RadSec = 2.0*(np.pi)/60.0

class ControllerInterface():
    """
    Define interface to a given controller using the avrSWAP array

    Methods:
    --------
    call_discon
    call_controller
    show_control_values

    Parameters:
    -----------
    lib_name : str
                name of compiled dynamic library containing controller, (.dll,.so,.dylib)

    """

    def __init__(self, lib_name, param_filename='DISCON.IN', **kwargs):
        """
        Setup the interface
        """
        self.lib_name = lib_name
        self.param_name = param_filename

        # Set default parameters
        # PARAMETERS
        self.DT = 0.1
        self.num_blade = 3
        self.char_buffer = 500
        self.avr_size = 500
        self.sim_name = 'simDEBUG'

        # Set kwargs, like DT
        for (k, w) in kwargs.items():
            try:
                setattr(self, k, w)
            except:
                pass

        self.init_discon()

    def init_discon(self):

        # Initialize
        self.pitch = 0
        self.torque = 0
        # -- discon
        self.discon = cdll.LoadLibrary(self.lib_name)
        self.avrSWAP = np.zeros(self.avr_size)

        # Define some avrSWAP parameters, NOTE: avrSWAP indices are offset by -1 from Fortran
        self.avrSWAP[2] = self.DT
        self.avrSWAP[60] = self.num_blade
        self.avrSWAP[19] = 1.0 # HARD CODE initial gen speed = 1 rad/s
        self.avrSWAP[20] = 1.0 # HARD CODE initial rot speed = 1 rad/s
        self.avrSWAP[82] = 0 # HARD CODE initial nacIMU = 0
        self.avrSWAP[26] = 10 # HARD CODE initial wind speed = 10 m/s
        
        # Blade pitch initial conditions
        self.avrSWAP[3]     = 0 * np.deg2rad(1)
        self.avrSWAP[32]    = 0 * np.deg2rad(1)
        self.avrSWAP[33]    = 0 * np.deg2rad(1)

        # Torque initial condition
        self.avrSWAP[22]    = 0


        # Code this as first casll
        self.avrSWAP[0] = 0

        # Put some values in
        self.avrSWAP[58] = self.char_buffer
        self.avrSWAP[49] = len(self.param_name)
        self.avrSWAP[50] = self.char_buffer
        self.avrSWAP[51] = self.char_buffer

        # Initialize DISCON and related
        self.aviFAIL = c_int32() # 1
        self.accINFILE = self.param_name.encode('utf-8')
        # self.avcOUTNAME = create_string_buffer(1000) # 'DEMO'.encode('utf-8')
        self.avcOUTNAME = (self.sim_name + '.RO.dbg').encode('utf-8')
        self.avcMSG = create_string_buffer(1000)
        self.discon.DISCON.argtypes = [POINTER(c_float), POINTER(c_int32), c_char_p, c_char_p, c_char_p] # (all defined by ctypes)

        # Run DISCON
        self.call_discon()

        # Code as not first run
        self.avrSWAP[0] = 1


    def call_discon(self):
        '''
        Call libdiscon.dll (or .so,.dylib,...)
        '''
        # Convert AVR swap to the c pointer
        c_float_p = POINTER(c_float)
        data = self.avrSWAP.astype(np.float32)
        p_data = data.ctypes.data_as(c_float_p)

        # Run DISCON
        self.discon.DISCON(p_data, byref(self.aviFAIL), self.accINFILE, self.avcOUTNAME, self.avcMSG)

        # Push back to avr swap
        self.avrSWAP = data


    def call_controller(self,t,dt,pitch,torque,genspeed,geneff,rotspeed,ws,NacIMU_FA_Acc=0):
        '''
        Runs the controller. Passes current turbine state to the controller, and returns control inputs back
        
        Parameters:
        -----------
        t: float
           time, (s)
        dt: float
            timestep, (s)
        pitch: float
               blade pitch, (rad)
        genspeed: float
                  generator speed, (rad/s)
        geneff: float
                  generator efficiency, (rad/s)
        rotspeed: float
                  rotor speed, (rad/s)
        ws: float
            wind speed, (m/s)
        NacIMU_FA_Acc : float
            nacelle IMU accel. in the nodding dir. , (m/s**2)
            default to 0 (fixed-bottom, simple 1-DOF sim does not include it, but OpenFAST linearizations do)
        '''

        # Add states to avr
        self.avrSWAP[1] = t
        self.avrSWAP[2] = dt
        self.avrSWAP[3] = pitch
        self.avrSWAP[32] = pitch
        self.avrSWAP[33] = pitch
        self.avrSWAP[14] = genspeed*torque*geneff
        self.avrSWAP[22] = torque
        self.avrSWAP[19] = genspeed
        self.avrSWAP[20] = rotspeed
        self.avrSWAP[26] = ws
        self.avrSWAP[82] = NacIMU_FA_Acc

        # call controller
        self.call_discon()

        # return controller states
        self.pitch = self.avrSWAP[41]
        self.torque = self.avrSWAP[46]

        return(self.torque,self.pitch)

    def show_control_values(self):
        '''
        Show control values - should be obvious
        '''
        print('Pitch',self.pitch)
        print('Torque',self.torque)

    def kill_discon(self):
        '''
        Unload the dylib from memory: https://github.com/bwoodsend/cslug/blob/master/cslug/_stdlib.py
        '''

        print('Shutting down {}'.format(self.lib_name))
        handle = self.discon._handle

        # Start copy here
        OS = platform.system()

        def null_free_dll(*spam):  # pragma: no cover
            pass

        # Try to find a good runtime library which is always available and contains
        # the standard library C functions such as malloc() or printf().
        # XXX: Keep chosen library names in sync with the table in `cslug/stdlib.py`.

        extra_libs = []

        if OS == "Windows":  # pragma: Windows
            _dlclose = ctypes.windll.kernel32.FreeLibrary
            dlclose = lambda handle: 0 if _dlclose(handle) else 1
            # There's some controversy as to whether this DLL is guaranteed to exist.
            # It always has so far but isn't documented. However, MinGW assumes that it
            # is so, should this DLL be removed, then we have much bigger problems than
            # just this line. There is also vcruntime140.dll which isn't a standard part
            # of the OS but is always shipped with Python so we can guarantee its
            # presence. But vcruntime140 contains only a tiny strict-subset of msvcrt.
            stdlib = ctypes.CDLL("msvcrt")

        elif OS == "Darwin":  # pragma: Darwin
            try:
                try:
                    # macOS 11 (Big Sur). Possibly also later macOS 10s.
                    stdlib = ctypes.CDLL("libc.dylib")
                except OSError:  # pragma: no cover
                    stdlib = ctypes.CDLL("libSystem")
            except OSError:  # pragma: no cover
                # Older macOSs. Not only is the name inconsistent but it's
                # not even in PATH.
                _stdlib = "/usr/lib/system/libsystem_c.dylib"
                if os.path.exists(_stdlib):
                    stdlib = ctypes.CDLL(_stdlib)
                else:
                    stdlib = None
            if stdlib is not None:  # pragma: no branch
                dlclose = stdlib.dlclose
            else:  # pragma: no cover
                # I hope this never happens.
                dlclose = null_free_dll

        elif OS == "Linux":  # pragma: Linux
            try:
                stdlib = ctypes.CDLL("")
            except OSError:  # pragma: no cover
                # Either Alpine Linux or Android.
                # Unfortunately, there doesn't seem to be any practical way
                # to tell them apart.
                stdlib = ctypes.CDLL("libc.so")

                # Android, like FreeBSD puts its math functions
                # in a dedicated `libm.so`.
                # The only way to know that this is not Alpine is to check if the math
                # functions are already available in `libc.so`.
                if not hasattr(stdlib, "sin"):
                    extra_libs.append(ctypes.CDLL("libm.so"))
            dlclose = stdlib.dlclose

        # End copy here
        dlclose.argtypes = [c_void_p]
        dlclose(handle)

        del self.discon



'''
Methods for finding robust gain schedules
'''

import numpy as np
import glob
import os
import openmdao.api as om
import matplotlib.pyplot as plt
import pandas as pd
import multiprocessing as mp
from ROSCO_toolbox import controller as ROSCO_controller
from ROSCO_toolbox import turbine as ROSCO_turbine
from ROSCO_toolbox.linear.linear_models import LinearTurbineModel
from ROSCO_toolbox.linear.lin_util import add_pcomp, smargin
from ROSCO_toolbox.inputs.validation import load_rosco_yaml
from ROSCO_toolbox.utilities import list_check

class RobustScheduling(om.ExplicitComponent):
    'Finding Robust gain schedules for pitch controllers in FOWTs'

    def initialize(self):
        self.options.declare('linturb_options')
        self.options.declare('ROSCO_options')

    def setup(self):
        # Options
        linturb_options = self.options['linturb_options']
        ROSCO_options = self.options['ROSCO_options']
        ROSCO_options['controller_params']['omega_pc'] = list_check(
            ROSCO_options['controller_params']['omega_pc'], return_bool=False)
        ROSCO_options['controller_params']['zeta_pc'] = list_check(
            ROSCO_options['controller_params']['zeta_pc'], return_bool=False)
        if list_check(ROSCO_options['controller_params']['omega_pc']) or \
                list_check(ROSCO_options['controller_params']['zeta_pc']):
            raise AttributeError(
                'Error: omega_pc and zeta_pc must be scalars for robust controller tuning.')

        # Load ROSCO Turbine and Controller
        if 'dict_inputs' in ROSCO_options.keys():  # Allow for turbine parameters to be passed in as a dictionary
            dict_inputs = ROSCO_options['dict_inputs']
            # Define turbine based on inputs
            self.turbine = type('', (), {})()
            self.turbine.v_min = float(dict_inputs['v_min'])
            self.turbine.J = float(dict_inputs['rotor_inertia'])
            self.turbine.rho = float(dict_inputs['rho'])
            self.turbine.rotor_radius = float(dict_inputs['R'])
            self.turbine.Ng = float(dict_inputs['gear_ratio'])
            # Incoming value already has gearbox eff included, so have to separate it out
            self.turbine.GenEff = float(
                dict_inputs['generator_efficiency']/dict_inputs['gearbox_efficiency']) * 100.
            self.turbine.GBoxEff = float(dict_inputs['gearbox_efficiency']) * 100.
            self.turbine.rated_rotor_speed = float(dict_inputs['rated_rotor_speed'])
            self.turbine.rated_power = float(dict_inputs['rated_power'])
            self.turbine.rated_torque = float(
                dict_inputs['rated_torque']) / self.turbine.Ng * float(dict_inputs['gearbox_efficiency'])
            self.turbine.v_rated = float(
                dict_inputs['rated_rotor_speed'])*float(dict_inputs['R']) / float(dict_inputs['tsr_operational'])
            self.turbine.v_min = float(dict_inputs['v_min'])
            self.turbine.v_max = float(dict_inputs['v_max'])
            self.turbine.max_pitch_rate = float(dict_inputs['max_pitch_rate'])
            self.turbine.TSR_operational = float(dict_inputs['tsr_operational'])
            self.turbine.max_torque_rate = float(dict_inputs['max_torque_rate'])
            self.turbine.TowerHt = float(dict_inputs['TowerHt'])
            self.turbine.Cp_table = np.squeeze(dict_inputs['Cp_table'])
            self.turbine.Ct_table = np.squeeze(dict_inputs['Ct_table'])
            self.turbine.Cq_table = np.squeeze(dict_inputs['Cq_table'])
            self.turbine.pitch_initial_rad = dict_inputs['pitch_vector']
            self.turbine.bld_edgewise_freq = float(dict_inputs['edge_freq'])
            self.turbine.TSR_initial = dict_inputs['tsr_vector']
            RotorPerformance = ROSCO_turbine.RotorPerformance
            self.turbine.Cp = RotorPerformance(
                self.turbine.Cp_table, self.turbine.pitch_initial_rad, self.turbine.TSR_initial)
            self.turbine.Ct = RotorPerformance(
                self.turbine.Ct_table, self.turbine.pitch_initial_rad, self.turbine.TSR_initial)
            self.turbine.Cq = RotorPerformance(
                self.turbine.Cq_table, self.turbine.pitch_initial_rad, self.turbine.TSR_initial)

            self.controller = ROSCO_controller.Controller(ROSCO_options['controller_params'])
            self.controller.tune_controller(self.turbine)
        else:
            self.controller, self.turbine = load_ROSCO(ROSCO_options['path_params'],
                                                       ROSCO_options['turbine_params'],
                                                       ROSCO_options['controller_params'])
        # Load linear turbine models and trim them
        self.linturb = load_linturb(
            linturb_options['linfile_path'], load_parallel=linturb_options['load_parallel'])
        self.linturb.trim_system(desInputs=['collective'], desOutputs=['RtSpeed'])

        # Inputs
        self.add_input('u_eval',    val=11.,  units='m/s',
                       desc='Wind speeds to evaluate gain schedule')
        self.add_input('omega',     val=0.1,  units='rad/s',   desc='Controller bandwidth')
        self.add_input('k_float',   val=self.controller.Kp_float,
                       units='s',       desc='Floating feedback gain')

        # Outputs
        self.add_output('sm',        val=0.0,                       desc='Stability Margin')
        self.add_output('omega_opt', val=0.01,      units='rad/s',
                        desc='Maximized controller bandwidth')

    def compute(self, inputs, outputs):
        k_float = inputs['k_float'][0]
        if k_float:
            linturb = add_pcomp(self.linturb, k_float)
        else:
            linturb = self.linturb

        self.controller.omega_pc = inputs['omega'][0]
        self.controller.tune_controller(self.turbine)
        sm = smargin(linturb, self.controller, inputs['u_eval'][0])

        omega = inputs['omega']

        # Outputs
        outputs['sm'] = sm
        outputs['omega_opt'] = omega


class rsched_driver():
    '''
    A driver for scheduling robust controllers
    '''

    def __init__(self, options):
        self.linturb_options = options['linturb_options']
        self.ROSCO_options = options['ROSCO_options']
        self.path_options = options['path_options']
        self.opt_options = options['opt_options']
        self.output_dir = options['path_options']['output_dir']
        self.output_name = options['path_options']['output_name']

        # Output options
        self.output_dir = self.path_options['output_dir']
        self.output_name = self.path_options['output_name']

        if 'levels' in self.opt_options.keys():
            self.opt_options['levels'] = self.opt_options['levels']
        else:
            if self.opt_options['driver'] == 'design_of_experiments':
                self.opt_options['levels'] = 20
            elif self.opt_options['driver'] == 'optimization':
                self.opt_options['levels'] = 10

        # Clarify up input sizes
        self.opt_options['omega'] = list_check(self.opt_options['omega'], return_bool=False)
        self.opt_options['k_float'] = list_check(self.opt_options['k_float'], return_bool=False)

    def setup(self):
        '''
        Setup the OpenMDAO problem
        '''
        # Initialize problem
        self.om_problem = om.Problem()

        # Add subsystem
        self.om_problem.model.add_subsystem('r_sched', RobustScheduling(linturb_options=self.linturb_options,
                                                                        ROSCO_options=self.ROSCO_options))

        if self.opt_options['driver'] == 'design_of_experiments':
            self.om_problem = self.init_doe(self.om_problem, levels=self.opt_options['levels'])
            if list_check(self.opt_options['windspeed']):
                if len(self.opt_options['windspeed']) == 1:
                    self.opt_options['windspeed'] = self.opt_options['windspeed'][0]
                else:
                    ValueError(
                        'Can only run design of experiments for a single opt_options["windspeed"]')
        elif self.opt_options['driver'] == 'optimization':
            self.om_problem = self.init_doe(self.om_problem, levels=self.opt_options['levels'])
        else:
            ValueError(
                "self.opt_options['driver'] must be either 'design_of_experiments' or 'optimization'.")

        # Add stability margin constraints
        self.om_problem.model.add_constraint(
            'r_sched.sm', lower=self.opt_options['stability_margin'])

        # Define objective
        self.om_problem.model.add_objective('r_sched.omega_opt', scaler=-1)

        # Setup OM Problem
        self.om_problem.setup()

        # Set constant values
        if not list_check(self.opt_options['omega']):
            self.om_problem.set_val('r_sched.omega', self.opt_options['omega'])
        if not list_check(self.opt_options['k_float']):
            self.om_problem.set_val('r_sched.k_float', self.opt_options['k_float'])

        # Designate specific problem objects, add design variables
        if self.opt_options['driver'] == 'design_of_experiments':
            self.om_doe = self.om_problem
            self.om_doe = self.add_dv(self.om_doe, ['omega', 'k_float'])
        if self.opt_options['driver'] == 'optimization':
            # self.om_doe = copy.deepcopy(self.om_problem)
            # self.om_doe = self.add_dv(self.om_doe, ['omega'])

            self.om_opt = self.om_problem
            self.om_opt = self.add_dv(self.om_opt, ['omega', 'k_float'])
            self.om_opt = self.init_optimization(self.om_opt)

    def execute(self):
        '''
        Execute OpenMDAO
        '''
        if self.opt_options['driver'] == 'design_of_experiments':
            u = self.opt_options['windspeed']
            self.om_doe.set_val('r_sched.u_eval', u)
            self.doe_logfile = os.path.join(
                self.output_dir, self.output_name + '.' + str(u) + ".doe.sql")
            self.om_doe.driver.add_recorder(om.SqliteRecorder(self.doe_logfile))
            self.om_doe.run_driver()

        elif self.opt_options['driver'] == 'optimization':
            self.omegas = []
            self.k_floats = []
            self.sms = []
            for u in self.opt_options['windspeed']:
                om0 = 0.1
                # Setup optimization
                self.om_opt.set_val('r_sched.u_eval', u)
                self.om_opt.set_val('r_sched.omega', om0)

                # Run optimization
                print('Finding ROSCO tuning parameters for u = {}, sm = {}'.format(
                    u, self.linturb_options['stability_margin']))
                opt_logfile = os.path.join(
                    self.output_dir, self.output_name + '.' + str(u) + ".opt.sql")
                self.om_opt = self.setup_recorder(self.om_opt, opt_logfile)
                self.om_opt.run_driver()
                self.om_opt.cleanup()

                # save values
                self.omegas.append(self.om_opt.get_val('r_sched.omega')[0])
                self.k_floats.append(self.om_opt.get_val('r_sched.k_float')[0])
                self.sms.append(self.om_opt.get_val('r_sched.sm')[0])

    def plot_schedule(self):
        fig, ax = plt.subplots(3, 1, constrained_layout=True, sharex=True)
        ax[0].plot(self.opt_options['windspeed'], self.omegas)
        ax[0].set_ylabel('omega_pc')
        ax[0].grid()
        ax[1].plot(self.opt_options['windspeed'], self.k_floats)
        ax[1].set_ylabel('k_float')
        ax[1].grid()
        ax[2].plot(self.opt_options['windspeed'], self.sms)
        ax[2].set_ylabel('stability margin')
        ax[2].set_xlabel('Wind Speed')
        ax[2].grid()
        plt.show()

    def add_dv(self, om_problem, opt_vars):
        '''add design variables'''

        if 'omega' in opt_vars and list_check(self.opt_options['omega']):
            om_problem.model.add_design_var(
                'r_sched.omega', lower=self.opt_options['omega'][0], upper=self.opt_options['omega'][1])

        if 'k_float' in opt_vars and list_check(self.opt_options['k_float']):
            om_problem.model.add_design_var(
                'r_sched.k_float', lower=self.opt_options['k_float'][0], upper=self.opt_options['k_float'][1], ref=100)

        # Make sure design variables are stored appropriately in OM problem
        om_problem.model._design_vars = om_problem.model._static_design_vars

        return om_problem

    def init_optimization(self, om_problem):
        'Initialize optimizer'
        om_problem.driver = om.ScipyOptimizeDriver()
        om_problem.driver.options['optimizer'] = 'SLSQP'
        om_problem.driver.options['tol'] = 1e-3
        om_problem.driver.options['maxiter'] = 20
        # om_problem.driver.options['debug_print'] = ['desvars', 'nl_cons']
        om_problem.model.approx_totals(method="fd", step=1e-1, form='central', step_calc='rel')

        return om_problem

    def init_doe(self, om_problem, levels=20):
        '''Initialize DOE driver'''
        om_problem.driver = om.DOEDriver(om.FullFactorialGenerator(levels=levels))
        # om_problem.driver = om.DOEDriver(om.UniformGenerator(num_samples=20))
        os.makedirs(self.output_dir, exist_ok=True)
        # om_problem.driver.options['run_parallel'] = True
        # om_problem.driver.options['procs_per_model'] = 1
        return om_problem

    @staticmethod
    def setup_recorder(problem, sql_filename):
        ''' Used to prevent memory issues with OM sqlite recorder'''
        recorder = om.SqliteRecorder(sql_filename)

        try:  # Try to remove previous recorder
            problem.driver._recorders.pop()
        except:  # Must be first pass or optimization run
            pass

        problem.driver.add_recorder(recorder)

        try:  # try to re-run recorder setup
            problem._setup_recording()
            problem.driver._setup_recording()
        except:  # Must be first pass
            pass

        return problem

    def post_doe(self, save_csv=False):
        '''Post process doe'''
        # write to file
        if save_csv:
            doe_outfile = '.'.join(self.doe_logfile.split('.')[:-1]) + '.csv'
        else:
            doe_outfile = None
        df = load_DOE(self.doe_logfile, outfile_name=doe_outfile)

        return df


def load_DOE(doe_logs, outfile_name=None):
    if isinstance(doe_logs, str):
        doe_logs = [doe_logs]

    if False:
        cores = mp.cpu_count()
        pool = mp.Pool(min(len(doe_logs), cores))

        # load sql file
        outdata = pool.map(load_OMsql, doe_logs)
        pool.close()
        pool.join()
    # no multiprocessing
    else:
        outdata = [load_OMsql(log) for log in doe_logs]

    collected_data = {}
    for data in outdata:
        for key in data.keys():
            if key not in collected_data.keys():
                collected_data[key] = []

            for key_idx, _ in enumerate(data[key]):
                if isinstance(data[key][key_idx], int):
                    collected_data[key].append(np.array(data[key][key_idx]))
                elif len(data[key][key_idx]) == 1:
                    try:
                        collected_data[key].append(np.array(data[key][key_idx][0]))
                    except:
                        collected_data[key].append(np.array(data[key][key_idx]))
                else:
                    collected_data[key].append(np.array(data[key][key_idx]))

    df = pd.DataFrame.from_dict(collected_data, dtype=float)

    if outfile_name:
        df.to_csv(outfile_name, index=False)
        print('Saved {}'.format(outfile_name))

    return df


def load_OMsql(log):
    print('loading {}'.format(log))
    cr = om.CaseReader(log)
    rec_data = {}
    driver_cases = cr.list_cases('driver')
    cases = cr.get_cases('driver')
    for case in cases:
        for key in case.outputs.keys():
            if key not in rec_data:
                rec_data[key] = []
            rec_data[key].append(case[key])

    return rec_data


def load_linturb(linfile_path, load_parallel=False):
    # Parse openfast linearization filenames
    filenames = glob.glob(os.path.join(linfile_path, '*.lin'))
    linfiles = [os.path.split(file)[1] for file in filenames]
    linroots = np.sort(np.unique([file.split('.')[0] for file in linfiles])).tolist()
    linfile_numbers = set([int(file.split('.')[1]) for file in linfiles])
    # Load linturb
    linturb = LinearTurbineModel(linfile_path, linroots,
                                 nlin=max(linfile_numbers), rm_hydro=True, load_parallel=load_parallel)

    return linturb

def load_ROSCO(path_params, turbine_params, controller_params):
    turbine = ROSCO_turbine.Turbine(turbine_params)
    controller = ROSCO_controller.Controller(controller_params)

    # Load turbine, tune controller
    turbine.load_from_fast(path_params['FAST_InputFile'], path_params['FAST_directory'])
    controller.tune_controller(turbine)

    return controller, turbine


if __name__ == '__main__':

    # Setup linear turbine paths
    linfile_path = os.path.join(os.path.dirname(os.path.dirname(
        os.path.dirname(os.path.abspath(__file__)))), 'Test_Cases', 'IEA-15-240-RWT-UMaineSemi', 'linearizations')
    load_parallel = True
    linturb_options = {'linfile_path': linfile_path,
                       'load_parallel': load_parallel}

    # ROSCO options
    parameter_filename = os.path.join(os.path.dirname(
        os.path.dirname(os.path.dirname(os.path.abspath(__file__)))), 'Tune_Cases', 'IEA15MW.yaml')
    inps = load_rosco_yaml(parameter_filename)
    path_params = inps['path_params']
    turbine_params = inps['turbine_params']
    controller_params = inps['controller_params']
    path_params['FAST_directory'] = os.path.join(os.path.dirname(
        os.path.dirname(os.path.dirname(os.path.abspath(__file__)))), 'Test_Cases', 'IEA-15-240-RWT-UMaineSemi')
    ROSCO_options = {
        'path_params': path_params,
        'turbine_params': turbine_params,
        'controller_params': controller_params
    }

    # Path options
    output_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'test_out')
    output_name = 'test'
    path_options = {'output_dir': output_dir,
                    'output_name': output_name
                    }

    # Scheduling options
    opt_options = {'driver': 'optimization',  # 'design_of_experiments',
                   'windspeed': [12, 13, 14],  # , 17, 18, 19, 20, 21, 22, 23, 24, 25],
                   'stability_margin': 0.1,
                   'omega': [0.05, 0.2],
                   'k_float': [0.0]}

    options = {}
    options['linturb_options'] = linturb_options
    options['ROSCO_options'] = ROSCO_options
    options['path_options'] = path_options
    options['opt_options'] = opt_options

    sd = rsched_driver(options)
    sd.setup()
    sd.execute()
    if opt_options['driver'] == 'design_of_experiments':
        sd.post_doe(save_csv=True)
    else:
        sd.plot_schedule()

if __name__ == '__main__':
    pass



# Copyright 2019 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy of the
# License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import numpy as np
import sys, os
import datetime
from scipy import interpolate, gradient, integrate
from ROSCO_toolbox.utilities import list_check

# Some useful constants
now = datetime.datetime.now()
pi = np.pi
rad2deg = np.rad2deg(1)
deg2rad = np.deg2rad(1)
rpm2RadSec = 2.0*(np.pi)/60.0
RadSec2rpm = 60/(2.0 * np.pi)

class Controller():
    """
    Class Controller used to calculate controller tunings parameters


    Methods:
    -------
    tune_controller

    Parameters:
    -----------
    controller_params: dict
                       Dictionary containing controller paramaters that need to be defined
    """

    def __init__(self, controller_params):
        ''' 
        Load controller tuning parameters from input dictionary
        '''

        print('-----------------------------------------------------------------------------')
        print('   Tuning a reference wind turbine controller using NREL\'s ROSCO toolbox    ')
        print('-----------------------------------------------------------------------------')

        # Controller Flags
        self.LoggingLevel       = controller_params['LoggingLevel']
        self.F_LPFType          = controller_params['F_LPFType']
        self.F_NotchType        = controller_params['F_NotchType']
        self.IPC_ControlMode    = controller_params['IPC_ControlMode']
        self.VS_ControlMode     = controller_params['VS_ControlMode']
        self.PC_ControlMode     = controller_params['PC_ControlMode']
        self.Y_ControlMode      = controller_params['Y_ControlMode']
        self.SS_Mode            = controller_params['SS_Mode']
        self.WE_Mode            = controller_params['WE_Mode']
        self.PS_Mode            = controller_params['PS_Mode']
        self.SD_Mode            = controller_params['SD_Mode']
        self.Fl_Mode            = controller_params['Fl_Mode']
        self.Flp_Mode           = controller_params['Flp_Mode']

        # Necessary parameters
        self.U_pc = list_check(controller_params['U_pc'], return_bool=False)
        self.zeta_pc = list_check(controller_params['zeta_pc'], return_bool=False)
        self.omega_pc = list_check(controller_params['omega_pc'], return_bool=False)
        self.zeta_vs = controller_params['zeta_vs']
        self.omega_vs = controller_params['omega_vs']
        self.interp_type = controller_params['interp_type']

        # Optional parameters with defaults
        self.min_pitch          = controller_params['min_pitch']
        self.max_pitch          = controller_params['max_pitch']
        self.vs_minspd          = controller_params['vs_minspd']
        self.ss_vsgain          = controller_params['ss_vsgain']
        self.ss_pcgain          = controller_params['ss_pcgain']
        self.ps_percent         = controller_params['ps_percent']
        self.sd_maxpit          = controller_params['sd_maxpit']
        self.WS_GS_n            = controller_params['WS_GS_n']
        self.PC_GS_n            = controller_params['PC_GS_n']
        self.flp_maxpit         = controller_params['flp_maxpit']
        self.Kp_ipc1p           = controller_params['IPC_Kp1p']
        self.Ki_ipc1p           = controller_params['IPC_Ki1p']
        self.Kp_ipc2p           = controller_params['IPC_Kp2p']
        self.Ki_ipc2p           = controller_params['IPC_Kp2p']

        #  Optional parameters without defaults
        if self.Flp_Mode > 0:
            try:
                self.flp_kp_norm = controller_params['flp_kp_norm']
                self.flp_tau     = controller_params['flp_tau']
            except:
                raise Exception(
                    'ROSCO_toolbox:controller: flp_kp_norm and flp_tau must be set if Flp_Mode > 0')

        if self.Fl_Mode > 0:
            try:
                self.twr_freq   = controller_params['twr_freq']
                self.ptfm_freq  = controller_params['ptfm_freq']
            except:
                raise Exception('ROSCO_toolbox:controller: twr_freq and ptfm_freq must be set if Fl_Mode > 0')

            # Kp_float direct setting
            if 'Kp_float' in controller_params:
                self.Kp_float = controller_params['Kp_float']
            else:
                self.Kp_float = 0

            self.tune_Fl = controller_params['tune_Fl']


        else:
            self.twr_freq   = 0
            self.ptfm_freq  = 0

        # Use critical damping if LPFType = 2
        if controller_params['F_LPFType'] == 2:
            self.F_LPFDamping = 0.7
        else:
            self.F_LPFDamping = 0.0

        # Filter parameters
        self.f_we_cornerfreq        = controller_params['filter_params']['f_we_cornerfreq']
        self.f_fl_highpassfreq      = controller_params['filter_params']['f_fl_highpassfreq']
        self.f_ss_cornerfreq        = controller_params['filter_params']['f_ss_cornerfreq']
        self.f_sd_cornerfreq        = controller_params['filter_params']['f_sd_cornerfreq']

        # Open loop parameters: set up and error catching
        self.OL_Mode            = int(controller_params['open_loop']['flag'])
        self.OL_Filename        = controller_params['open_loop']['filename']
        self.OL_Ind_Breakpoint  = self.OL_Ind_BldPitch = self.OL_Ind_GenTq = self.OL_Ind_YawRate = 0
        
        if self.OL_Mode:
            ol_params               = controller_params['open_loop']
            
                
            self.OL_Ind_Breakpoint  = ol_params['OL_Ind_Breakpoint']
            self.OL_Ind_BldPitch    = ol_params['OL_Ind_BldPitch']
            self.OL_Ind_GenTq       = ol_params['OL_Ind_GenTq']
            self.OL_Ind_YawRate     = ol_params['OL_Ind_YawRate']

            # Check that file exists because we won't write it
            if not os.path.exists(self.OL_Filename):
                raise Exception(f'Open-loop control set up, but the open loop file {self.OL_Filename} does not exist')
            

        # Save controller_params for later (direct passthrough)
        self.controller_params = controller_params

        # Error checking: number of breakpoints
        if self.WS_GS_n <= self.PC_GS_n:
            raise Exception('Number of WS breakpoints is not greater than pitch control breakpoints')

        # Error checking: pitch controller inputs
        if list_check(self.U_pc) and \
            (list_check(self.omega_pc) or list_check(self.zeta_pc)) and \
                not len(self.U_pc) == len(self.omega_pc) == len(self.zeta_pc):
            raise Exception(
                'U_pc, omega_pc, and zeta_pc are all list-like and are not of equal length')
        

    def tune_controller(self, turbine):
        """
        Given a turbine model, tune a controller based on the NREL generic controller tuning process

        Parameters:
        -----------
        turbine : class
                  Turbine class containing necessary turbine information to accurately tune the controller. 
        """
        # -------------Load Parameters ------------- #
        # Re-define Turbine Parameters for shorthand
        J = turbine.J                           # Total rotor inertial (kg-m^2) 
        rho = turbine.rho                       # Air density (kg/m^3)
        R = turbine.rotor_radius                    # Rotor radius (m)
        Ar = np.pi*R**2                         # Rotor area (m^2)
        Ng = turbine.Ng                         # Gearbox ratio (-)
        rated_rotor_speed = turbine.rated_rotor_speed               # Rated rotor speed (rad/s)


        # -------------Define Operation Points ------------- #
        TSR_rated = rated_rotor_speed*R/turbine.v_rated  # TSR at rated

        # separate wind speeds by operation regions
        # add one to above rated because we don't use rated in the pitch control gain scheduling
        v_below_rated = np.linspace(turbine.v_min,turbine.v_rated, num=self.WS_GS_n-self.PC_GS_n)[:-1]             # below rated
        v_above_rated = np.linspace(turbine.v_rated,turbine.v_max, num=self.PC_GS_n+1)             # above rated
        v = np.concatenate((v_below_rated, v_above_rated))

        # separate TSRs by operations regions
        TSR_below_rated = [min(turbine.TSR_operational, rated_rotor_speed*R/v) for v in v_below_rated] # below rated     
        TSR_above_rated = rated_rotor_speed*R/v_above_rated                   # above rated
        # TSR_below_rated = np.minimum(np.max(TSR_above_rated), TSR_below_rated)
        TSR_op = np.concatenate((TSR_below_rated, TSR_above_rated))   # operational TSRs

        # Find expected operational Cp values
        Cp_above_rated = turbine.Cp.interp_surface(0,TSR_above_rated[0])             # Cp during rated operation (not optimal). Assumes cut-in bld pitch to be 0
        Cp_op_br = np.ones(len(v_below_rated)) * turbine.Cp.max              # below rated
        Cp_op_ar = Cp_above_rated * (TSR_above_rated/TSR_rated)**3           # above rated
        Cp_op = np.concatenate((Cp_op_br, Cp_op_ar))                # operational CPs to linearize around
        pitch_initial_rad = turbine.pitch_initial_rad
        TSR_initial = turbine.TSR_initial

        # initialize variables
        pitch_op    = np.empty(len(TSR_op))
        dCp_beta    = np.empty(len(TSR_op))
        dCp_TSR     = np.empty(len(TSR_op))
        dCt_beta    = np.empty(len(TSR_op))
        dCt_TSR     = np.empty(len(TSR_op))
        Ct_op       = np.empty(len(TSR_op))

        # ------------- Find Linearized State "Matrices" ------------- #
        for i in range(len(TSR_op)):
            # Find pitch angle as a function of expected operating CP for each TSR
            Cp_TSR = np.ndarray.flatten(turbine.Cp.interp_surface(turbine.pitch_initial_rad, TSR_op[i]))     # all Cp values for a given tsr
            Cp_maxidx = Cp_TSR.argmax()    
            Cp_op[i] = np.clip(Cp_op[i], np.min(Cp_TSR[Cp_maxidx:]), np.max(Cp_TSR[Cp_maxidx:])) # saturate Cp values to be on Cp surface                                                             # Find maximum Cp value for this TSR
            f_cp_pitch = interpolate.interp1d(Cp_TSR[Cp_maxidx:],pitch_initial_rad[Cp_maxidx:])         # interpolate function for Cp(tsr) values
            # expected operation blade pitch values
            if v[i] <= turbine.v_rated and isinstance(self.min_pitch, float): # Below rated & defined min_pitch
                pitch_op[i] = min(self.min_pitch, f_cp_pitch(Cp_op[i]))
            elif isinstance(self.min_pitch, float):
                pitch_op[i] = max(self.min_pitch, f_cp_pitch(Cp_op[i]))             
            else:
                pitch_op[i] = f_cp_pitch(Cp_op[i])     

            dCp_beta[i], dCp_TSR[i] = turbine.Cp.interp_gradient(pitch_op[i],TSR_op[i])       # gradients of Cp surface in Beta and TSR directions
            dCt_beta[i], dCt_TSR[i] = turbine.Ct.interp_gradient(pitch_op[i],TSR_op[i])       # gradients of Cp surface in Beta and TSR directions
        
            # Thrust
            Ct_TSR      = np.ndarray.flatten(turbine.Ct.interp_surface(turbine.pitch_initial_rad, TSR_op[i]))     # all Cp values for a given tsr
            f_ct        = interpolate.interp1d(pitch_initial_rad,Ct_TSR)
            Ct_op[i]    = f_ct(pitch_op[i])
            Ct_op[i]    = np.clip(Ct_op[i], np.min(Ct_TSR), np.max(Ct_TSR))        # saturate Ct values to be on Ct surface


        # Define minimum pitch saturation to be at Cp-maximizing pitch angle if not specifically defined
        if not isinstance(self.min_pitch, float):
            self.min_pitch = pitch_op[0]

        # Full Cx surface gradients
        dCp_dbeta   = dCp_beta/np.diff(pitch_initial_rad)[0]
        dCp_dTSR    = dCp_TSR/np.diff(TSR_initial)[0]
        dCt_dbeta   = dCt_beta/np.diff(pitch_initial_rad)[0]
        dCt_dTSR    = dCt_TSR/np.diff(TSR_initial)[0]
        
        # Linearized system derivatives
        dtau_dbeta      = Ng/2*rho*Ar*R*(1/TSR_op)*dCp_dbeta*v**2
        dtau_dlambda    = Ng/2*rho*Ar*R*v**2*(1/(TSR_op**2))*(dCp_dTSR*TSR_op - Cp_op)
        dlambda_domega  = R/v/Ng
        dtau_domega     = dtau_dlambda*dlambda_domega

        dlambda_dv      = -(TSR_op/v)

        Pi_beta         = 1/2 * rho * Ar * v**2 * dCt_dbeta
        Pi_omega        = 1/2 * rho * Ar * R * v * dCt_dTSR
        Pi_wind         = 1/2 * rho * Ar * v**2 * dCt_dTSR * dlambda_dv + rho * Ar * v * Ct_op

        # Second order system coefficients
        if self.VS_ControlMode in [0,2]: # Constant torque above rated
            A = dtau_domega/J
        else:                            # Constant power above rated
            A = dtau_domega/J 
            A[-len(v_above_rated)+1:] += Ng**2/J * turbine.rated_power/(Ng**2*rated_rotor_speed**2)
        B_tau = -Ng**2/J              # Torque input  
        B_beta = dtau_dbeta/J         # Blade pitch input 

        # Wind Disturbance Input
        dtau_dv = (0.5 * rho * Ar * 1/rated_rotor_speed) * (dCp_dTSR*dlambda_dv*v**3 + Cp_op*3*v**2) 
        B_wind = dtau_dv/J # wind speed input - currently unused 


        # separate and define below and above rated parameters
        A_vs = A[0:len(v_below_rated)]          # below rated
        A_pc = A[-len(v_above_rated)+1:]     # above rated
        B_tau = B_tau * np.ones(len(v))

        # Resample omega_ and zeta_pc at above rated wind speeds
        if not list_check(self.omega_pc) and not list_check(self.zeta_pc):
            self.omega_pc_U = self.omega_pc * np.ones(len(v_above_rated[1:]))
            self.zeta_pc_U = self.zeta_pc * np.ones(len(v_above_rated[1:]))
        else:
            if self.interp_type == 'sigma':  # sigma interpolation
                self.omega_pc_U = multi_sigma(v_above_rated[1:], self.U_pc, self.omega_pc)
                self.zeta_pc_U = multi_sigma(v_above_rated[1:], self.U_pc, self.zeta_pc)
            else:   # standard scipy interpolation types
                interp_omega = interpolate.interp1d(self.U_pc, self.omega_pc, kind=self.interp_type, bounds_error=False, fill_value='extrapolate')
                interp_zeta = interpolate.interp1d(self.U_pc, self.zeta_pc, kind=self.interp_type, bounds_error=False, fill_value='extrapolate')
                self.omega_pc_U = interp_omega(v_above_rated[1:])
                self.zeta_pc_U = interp_zeta(v_above_rated[1:])

        # -- Find gain schedule --
        self.pc_gain_schedule = ControllerTypes()
        self.pc_gain_schedule.second_order_PI(self.zeta_pc_U, self.omega_pc_U,A_pc,B_beta[-len(v_above_rated)+1:],linearize=True,v=v_above_rated[1:])        
        self.vs_gain_schedule = ControllerTypes()
        self.vs_gain_schedule.second_order_PI(self.zeta_vs, self.omega_vs,A_vs,B_tau[0:len(v_below_rated)],linearize=False,v=v_below_rated)

        # -- Find K for Komega_g^2 --
        self.vs_rgn2K = (pi*rho*R**5.0 * turbine.Cp.max) / (2.0 * turbine.Cp.TSR_opt**3 * Ng**3)/ (turbine.GBoxEff/100)
        self.vs_refspd = min(turbine.TSR_operational * turbine.v_rated/R, turbine.rated_rotor_speed) * Ng

        # -- Define some setpoints --
        # minimum rotor speed saturation limits
        if self.vs_minspd:
            self.vs_minspd = np.maximum(self.vs_minspd, (turbine.TSR_operational * turbine.v_min / turbine.rotor_radius))
        else: 
            self.vs_minspd = (turbine.TSR_operational * turbine.v_min / turbine.rotor_radius)
        self.pc_minspd = self.vs_minspd

        # max pitch angle for shutdown
        if self.sd_maxpit:
            self.sd_maxpit = self.sd_maxpit
        else:
            self.sd_maxpit = pitch_op[-1]

        # Store some variables
        self.v              = v                                  # Wind speed (m/s)
        self.v_below_rated  = v_below_rated
        self.pitch_op       = pitch_op
        self.pitch_op_pc    = pitch_op[-len(v_above_rated)+1:]
        self.TSR_op         = TSR_op
        self.A              = A 
        self.B_beta         = B_beta
        self.B_tau          = B_tau
        self.B_wind         = B_wind
        self.omega_op       = np.maximum(np.minimum(turbine.rated_rotor_speed, TSR_op*v/R), self.vs_minspd)
        self.Pi_omega       = Pi_omega
        self.Pi_beta        = Pi_beta
        self.Pi_wind        = Pi_wind

        # - Might want these to debug -
        self.Cp_op = Cp_op

        # --- Minimum pitch saturation ---
        self.ps_min_bld_pitch = np.ones(len(self.pitch_op)) * self.min_pitch
        self.ps = ControllerBlocks()

        if self.PS_Mode == 1:  # Peak Shaving
            self.ps.peak_shaving(self, turbine)
        elif self.PS_Mode == 2: # Cp-maximizing minimum pitch saturation
            self.ps.min_pitch_saturation(self,turbine)
        elif self.PS_Mode == 3: # Peak shaving and Cp-maximizing minimum pitch saturation
            self.ps.peak_shaving(self, turbine)
            self.ps.min_pitch_saturation(self,turbine)

        # --- Floating feedback term ---
        if self.Fl_Mode >= 1: # Floating feedback
            # If we haven't set Kp_float as a control parameter
            if self.tune_Fl:
                Kp_float = (dtau_dv/dtau_dbeta) * Ng 
                if self.Fl_Mode == 2:
                    Kp_float *= turbine.TowerHt      
                f_kp     = interpolate.interp1d(v,Kp_float)
                self.Kp_float = f_kp(turbine.v_rated * (1.05))   # get Kp at v_rated + 0.5 m/s

            # Turn on the notch filter if floating
            self.F_NotchType = 2
            
            # And check for .yaml input inconsistencies
            if self.twr_freq == 0.0 or self.ptfm_freq == 0.0:
                print('WARNING: twr_freq and ptfm_freq should be defined for floating turbine control!!')
            
        else:
            self.Kp_float = 0.0
        
        # Flap actuation 
        if self.Flp_Mode >= 1:
            self.flp_angle = 0.0
            try:
                self.tune_flap_controller(turbine)
            except AttributeError:
                print('ERROR: If Flp_Mode > 0, you need to have blade information loaded in the turbine object.')
                raise
            except UnboundLocalError:
                print('ERROR: You are attempting to tune a flap controller for a blade without flaps!')
                raise
        else:
            self.flp_angle = 0.0
            self.Ki_flap = np.array([0.0])
            self.Kp_flap = np.array([0.0])

        # --- Set up filters ---
        self.f_lpf_cornerfreq = turbine.bld_edgewise_freq / 4

        # --- Direct input passthrough ---
        if 'f_lpf_cornerfreq' in self.controller_params['filter_params']:
            self.f_lpf_cornerfreq = self.controller_params['filter_params']['f_lpf_cornerfreq']

    def tune_flap_controller(self,turbine):
        '''
        Tune controller for distributed aerodynamic control

        Parameters:
        -----------
        turbine : class
                  Turbine class containing necessary turbine information to accurately tune the controller. 
        '''
        # Find blade aerodynamic coefficients
        v_rel = []
        phi_vec = []
        alpha=[]
        for i, _ in enumerate(self.v):
            turbine.cc_rotor.induction_inflow=True
            # Axial and tangential inductions
            try: 
                a, ap, alpha0, cl, cd = turbine.cc_rotor.distributedAeroLoads(
                                                self.v[i], self.omega_op[i], self.pitch_op[i], 0.0)
            except ValueError:
                loads, derivs = turbine.cc_rotor.distributedAeroLoads(
                                                self.v[i], self.omega_op[i], self.pitch_op[i], 0.0)
                a = loads['a']
                ap = loads['ap']
                alpha0 = loads['alpha']
                cl = loads['Cl']
                cd = loads['Cd']
                 
            # Relative windspeed
            v_rel.append([np.sqrt(self.v[i]**2*(1-a)**2 + self.omega_op[i]**2*turbine.span**2*(1-ap)**2)])
            # Inflow wind direction
            phi_vec.append(self.pitch_op[i] + turbine.twist*deg2rad)

        # Lift and drag coefficients
        num_af = len(turbine.af_data) # number of airfoils
        Cl0 = np.zeros(num_af)
        Cd0 = np.zeros(num_af)
        Clp = np.zeros(num_af)
        Cdp = np.zeros(num_af)
        Clm = np.zeros(num_af)
        Cdm = np.zeros(num_af)
        
        for i,section in enumerate(turbine.af_data):
            # assume airfoil section as AOA of zero for slope calculations - for now
            a0_ind = section[0]['Alpha'].index(np.min(np.abs(section[0]['Alpha'])))
            # Coefficients 
            if section[0]['NumTabs'] == 3:  # sections with 3 flaps
                Clm[i,] = section[0]['Cl'][a0_ind]
                Cdm[i,] = section[0]['Cd'][a0_ind]
                Cl0[i,] = section[1]['Cl'][a0_ind]
                Cd0[i,] = section[1]['Cd'][a0_ind]
                Clp[i,] = section[2]['Cl'][a0_ind]
                Cdp[i,] = section[2]['Cd'][a0_ind]
                Ctrl_flp = float(section[2]['Ctrl'])
            else:                           # sections without 3 flaps
                Cl0[i,] = Clp[i,] = Clm[i,] = section[0]['Cl'][a0_ind]
                Cd0[i,] = Cdp[i,] = Cdm[i,] = section[0]['Cd'][a0_ind]
                Ctrl = float(section[0]['Ctrl'])

        # Find slopes
        Kcl = (Clp - Cl0)/( (Ctrl_flp-Ctrl)*deg2rad )
        Kcd = (Cdp - Cd0)/( (Ctrl_flp-Ctrl)*deg2rad )

        # Find integrated constants
        self.kappa = np.zeros(len(v_rel))
        C1 = np.zeros(len(v_rel))
        C2 = np.zeros(len(v_rel))
        for i, (v_sec,phi) in enumerate(zip(v_rel, phi_vec)):
            C1[i] = integrate.trapz(0.5 * turbine.rho * turbine.chord * v_sec[0]**2 * turbine.span * Kcl * np.cos(phi))
            C2[i] = integrate.trapz(0.5 * turbine.rho * turbine.chord * v_sec[0]**2 * turbine.span * Kcd * np.sin(phi))
            self.kappa[i]=C1[i]+C2[i]

        # PI Gains
        if (self.flp_kp_norm == 0 or self.flp_tau == 0) or (not self.flp_kp_norm or not self.flp_tau):
            raise ValueError('flp_kp_norm and flp_tau must be nonzero for Flp_Mode >= 1')

        self.Kp_flap = self.flp_kp_norm / self.kappa
        self.Ki_flap = self.flp_kp_norm / self.kappa / self.flp_tau

class ControllerBlocks():
    '''
    Class ControllerBlocks defines tuning parameters for additional controller features or "blocks"

    Methods:
    --------
    peak_shaving

    '''
    def __init__(self):
        pass
    
    def peak_shaving(self,controller, turbine):
        ''' 
        Define minimum blade pitch angle for peak shaving routine based on a maximum allowable thrust 

        Parameters:
        -----------
        controller: class
                    Controller class containing controller operational information
        turbine: class
                 Turbine class containing necessary wind turbine information for controller tuning
        '''

        # Re-define Turbine Parameters for shorthand
        J = turbine.J                           # Total rotor inertial (kg-m^2) 
        rho = turbine.rho                       # Air density (kg/m^3)
        R = turbine.rotor_radius                    # Rotor radius (m)
        A = np.pi*R**2                         # Rotor area (m^2)
        Ng = turbine.Ng                         # Gearbox ratio (-)
        rated_rotor_speed = turbine.rated_rotor_speed               # Rated rotor speed (rad/s)

        # Initialize some arrays
        Ct_op = np.empty(len(controller.TSR_op),dtype='float64')
        Ct_max = np.empty(len(controller.TSR_op),dtype='float64')
        beta_min = np.empty(len(controller.TSR_op),dtype='float64')
        # Find unshaved rotor thurst coefficients and associated rotor thrusts
        # for i in len(controller.TSR_op):
        for i in range(len(controller.TSR_op)):
            Ct_op[i] = turbine.Ct.interp_surface(controller.pitch_op[i],controller.TSR_op[i])
            T = 0.5 * rho * A * controller.v**2 * Ct_op

        # Define minimum max thrust and initialize pitch_min
        Tmax = controller.ps_percent * np.max(T)
        pitch_min = np.ones(len(controller.pitch_op)) * controller.min_pitch

        # Modify pitch_min if max thrust exceeds limits
        for i in range(len(controller.TSR_op)):
            # Find Ct values for operational TSR
            # Ct_tsr = turbine.Ct.interp_surface(turbine.pitch_initial_rad, controller.TSR_op[i])
            Ct_tsr = turbine.Ct.interp_surface(turbine.pitch_initial_rad,controller.TSR_op[i])
            # Define max Ct values
            Ct_max[i] = Tmax/(0.5 * rho * A * controller.v[i]**2)
            if T[i] > Tmax:
                Ct_op[i] = Ct_max[i]
            else:
                Ct_max[i] = np.minimum( np.max(Ct_tsr), Ct_max[i])
            # Define minimum pitch angle
            f_pitch_min = interpolate.interp1d(Ct_tsr, turbine.pitch_initial_rad, kind='linear', bounds_error=False, fill_value=(turbine.pitch_initial_rad[0],turbine.pitch_initial_rad[-1]))
            pitch_min[i] = max(controller.min_pitch, f_pitch_min(Ct_max[i]))

        controller.ps_min_bld_pitch = pitch_min

        # save some outputs for analysis or future work
        self.Tshaved = 0.5 * rho * A * controller.v**2 * Ct_op
        self.pitch_min = pitch_min
        self.v = controller.v
        self.Ct_max = Ct_max
        self.Ct_op = Ct_op
        self.T = T

    def min_pitch_saturation(self, controller, turbine):
        
        # Find TSR associated with minimum rotor speed
        TSR_at_minspeed = (controller.pc_minspd) * turbine.rotor_radius / controller.v_below_rated
        for i in range(len(TSR_at_minspeed)):
            if TSR_at_minspeed[i] > controller.TSR_op[i]:
                controller.TSR_op[i] = TSR_at_minspeed[i]
        
                # Initialize some arrays
                Cp_op = np.empty(len(turbine.pitch_initial_rad),dtype='float64')
                min_pitch = np.empty(len(TSR_at_minspeed),dtype='float64')
                
        
                # Find Cp-maximizing minimum pitch schedule
                # Find Cp coefficients at below-rated tip speed ratios
                Cp_op = turbine.Cp.interp_surface(turbine.pitch_initial_rad,TSR_at_minspeed[i])
                Cp_max = max(Cp_op)
                # f_pitch_min = interpolate.interp1d(Cp_op, -turbine.pitch_initial_rad, kind='quadratic', bounds_error=False, fill_value=(turbine.pitch_initial_rad[0],turbine.pitch_initial_rad[-1]))
                f_pitch_min = interpolate.interp1d(turbine.pitch_initial_rad, -Cp_op, kind='quadratic', bounds_error=False, fill_value=(turbine.pitch_initial_rad[0],turbine.pitch_initial_rad[-1]))
                from scipy import optimize
                res = optimize.minimize(f_pitch_min, 0.0)
                min_pitch[i] = res.x[0]
                # min_pitch[i] = f_pitch_min(Cp_max)
                
                # modify existing minimum pitch schedule
                controller.ps_min_bld_pitch[i] = np.maximum(controller.ps_min_bld_pitch[i], min_pitch[i])

                # Save Cp_op
                controller.Cp_op[i] = -res.fun
            else:
                return


class ControllerTypes():
    '''
    Class ControllerTypes used to define any types of controllers that can be tuned. 
        Generally, calculates gains based on some pre-defined tuning parameters. 

    Methods:
    --------
    second_order_PI
    '''
    def __init__(self):
        pass

    def second_order_PI(self,zeta,om_n,A,B,linearize=False,v=None):
        '''
        Define proportional integral gain schedule for a closed
            loop system with a standard second-order form.

        Parameters:
        -----------
        zeta : list of floats (-)
               Desired damping ratio with breakpoints at v
        om_n : list of floats (rad/s)
               Desired natural frequency with breakpoints at v
        A : array_like (1/s)
            Plant poles (state transition matrix)
        B : array_like (varies)
            Plant numerators (input matrix)
        linearize : bool, optional
                    If 'True', find a gain scheduled based on a linearized plant.
        v : array_like (m/s)
            Wind speeds for linearized plant model, if desired. 
        '''
        # Linearize system coefficients w.r.t. wind speed if desired
        if linearize:
            pA = np.polyfit(v,A,1)
            pB = np.polyfit(v,B,1)
            A = pA[0]*v + pA[1]
            B = pB[0]*v + pB[1]

        # Calculate gain schedule
        self.Kp = 1/B * (2*zeta*om_n + A)
        self.Ki = om_n**2/B           

class OpenLoopControl(object):
    '''
    Open loop controls for
        - blade_pitch
        - generator_torque
        - nacelle_yaw and nacelle_yaw_rate

    Please see Examples/example_14.py for an example on how to use this class.

    '''

    def __init__(self, **kwargs):
        self.dt     = 0.05
        self.t_max  = 200

        # Optional population class attributes from key word arguments
        for (k, w) in kwargs.items():
            try:
                setattr(self, k, w)
            except:
                pass

        self.ol_timeseries = {}
        self.ol_timeseries['time'] = np.arange(0,self.t_max,self.dt)

        self.allowed_controls = ['blade_pitch','generator_torque','nacelle_yaw','nacelle_yaw_rate']

        
    def const_timeseries(self,control,value):
        self.ol_timeseries[control] = value * np.ones(len(self.ol_timeseries['time']))
        

    def interp_timeseries(self,control,breakpoints,values,method='sigma'):

        # Error checking
        if not list_check(breakpoints) or len(breakpoints) == 1:
            raise Exception('Open loop breakpoints are not a list with length greater than 1')

        if not list_check(values) or len(values) == 1:
            raise Exception('Open loop values are not a list with length greater than 1')

        if len(breakpoints) != len(values):
            raise Exception('Open loop breakpoints and values do not have the same length')

        if control not in self.allowed_controls:
            raise Exception(f'Open loop control of {control} is not allowed')

        else:
            # Finally interpolate
            if method == 'sigma':
                self.ol_timeseries[control] = multi_sigma(self.ol_timeseries['time'],breakpoints,values)
            
            elif method == 'linear':
                interp_fcn = interpolate.interp1d(breakpoints,values,fill_value=values[-1],bounds_error=False)
                self.ol_timeseries[control] = interp_fcn(self.ol_timeseries['time'])

            elif method == 'cubic':
                interp_fcn = interpolate.interp1d(breakpoints,values,kind='cubic',fill_value=values[-1],bounds_error=False)
                self.ol_timeseries[control] = interp_fcn(self.ol_timeseries['time'])

            else:
                raise Exception(f'Open loop interpolation method {method} not supported')

        if control == 'nacelle_yaw':
            self.compute_yaw_rate()


    def sine_timeseries(self,control,amplitude,period):
        
        if period <= 0:
            raise Exception('Open loop sine input period is <= 0')

        if control not in self.allowed_controls:
            raise Exception(f'Open loop control of {control} is not allowed')
        else:
            self.ol_timeseries[control] = amplitude * np.sin(2 * np.pi *  self.ol_timeseries['time'] / period)

        if control == 'nacelle_yaw':
            self.compute_yaw_rate()

    def compute_yaw_rate(self):
        self.ol_timeseries['nacelle_yaw_rate'] = np.concatenate(([0],np.diff(self.ol_timeseries['nacelle_yaw'])))/self.dt

    def plot_timeseries(self):
        '''
        Debugging script for showing open loop timeseries
        '''
        import matplotlib.pyplot as plt
        fig, ax = plt.subplots(len(self.ol_timeseries)-1,1)
        i_ax = -1
        for ol_input in self.ol_timeseries:
            if ol_input != 'time':
                i_ax += 1
                if len(self.ol_timeseries)-1 == 1:
                    ax.plot(self.ol_timeseries['time'],self.ol_timeseries[ol_input])
                    ax.set_ylabel(ol_input)
                else:
                    ax[i_ax].plot(self.ol_timeseries['time'],self.ol_timeseries[ol_input])
                    ax[i_ax].set_ylabel(ol_input)
        return fig, ax

    def write_input(self,ol_filename):
        ''' 
        Write open loop control input
        Return open_loop dict for control params
        '''

        ol_timeseries = self.ol_timeseries

        OL_Ind_Breakpoint = 1
        OL_Ind_BldPitch = OL_Ind_GenTq = OL_Ind_YawRate = 0

        self.OL_Ind_Breakpoint = 1
        ol_index_counter = 2   # start input index at 2
        
        if 'time' in ol_timeseries:
            ol_control_array = ol_timeseries['time']

        else:
            raise Exception('WARNING: no time index for open loop control.  This is only index currently supported')

        if 'blade_pitch' in ol_timeseries:
            OL_Ind_BldPitch = ol_index_counter
            ol_index_counter += 1
            
            ol_control_array = np.c_[ol_control_array,ol_timeseries['blade_pitch']]

        if 'generator_torque' in ol_timeseries:
            OL_Ind_GenTq = ol_index_counter
            ol_index_counter += 1
            
            ol_control_array = np.c_[ol_control_array,ol_timeseries['generator_torque']]

        if 'nacelle_yaw_rate' in ol_timeseries:
            OL_Ind_YawRate = ol_index_counter
            ol_index_counter += 1

            ol_control_array = np.c_[ol_control_array,ol_timeseries['nacelle_yaw_rate']]

        if 'nacelle_yaw' in ol_timeseries and 'nacelle_yaw_rate' not in ol_timeseries:
            raise Exception('nacelle_yaw is in ol_timeseries and nacelle_yaw_rate is not.  ROSCO can only command yaw rate. Use compute_yaw_rate() to convert.')

        # Open file
        if not os.path.exists(os.path.dirname(os.path.abspath(ol_filename))):
            os.makedirs(os.path.dirname(os.path.abspath(ol_filename)))
        
        with open(ol_filename,'w') as f:
            # Write header
            header_line = '!\tTime'
            unit_line   = '!\t(sec.)'
            if OL_Ind_BldPitch:
                header_line += '\t\tBldPitch'
                unit_line   += '\t\t(rad.)'

            if OL_Ind_GenTq:
                header_line += '\t\tGenTq'
                unit_line   += '\t\t(Nm)'

            if OL_Ind_YawRate:
                header_line += '\t\tYawRate'
                unit_line   += '\t\t(rad/s)'

            header_line += '\n'
            unit_line   += '\n'

            f.write(header_line)
            f.write(unit_line)

            # Write lines
            for ol_line in ol_control_array:
                line = ''.join(['{:<10.8f}\t'.format(val) for val in ol_line]) + '\n'
                f.write(line)

        # Output open_loop dict for control params
        open_loop = {}
        open_loop['flag']               = True
        open_loop['filename']           = ol_filename
        open_loop['OL_Ind_Breakpoint']  = OL_Ind_Breakpoint
        open_loop['OL_Ind_BldPitch']    = OL_Ind_BldPitch
        open_loop['OL_Ind_GenTq']       = OL_Ind_GenTq
        open_loop['OL_Ind_YawRate']     = OL_Ind_YawRate

        return open_loop


# helper functions

def sigma(tt,t0,t1,y0=0,y1=1):
    ''' 
    generates timeseries for a smooth transition from y0 to y1 from x0 to x1

    inputs: tt - time indices
            t0 - start time
            t1 - end time
            y0 - start output
            y1 - end output

    outputs: yy - output timeseries corresponding to tt
    '''

    a3 = 2/(t0-t1)**3
    a2 = -3*(t0+t1)/(t0-t1)**3
    a1 = 6*t1*t0/(t0-t1)**3
    a0 = (t0-3*t1)*t0**2/(t0-t1)**3

    a = np.array([a3,a2,a1,a0])  

    T = np.vander(tt,N=4)       # vandermonde matrix

    ss = T @ a.T                # base sigma

    yy = (y1-y0) * ss + y0      # scale and offset

    return yy


def multi_sigma(xx,x_bp,y_bp):
    '''
    Make a sigma interpolation with multiple breakpoints

    Parameters:
    -----------
    xx : list of floats (-)
            new sample points
    x_bp : list of floats (-)
            breakpoints
    y_bp : list of floats (-)
            function value at breakpoints

    '''
    yy = np.zeros_like(xx)

    # interpolate sigma functions between all breakpoints
    for i_sigma in range(0,len(x_bp)-1):
        ind_i       = (xx >= x_bp[i_sigma]) & (xx < x_bp[i_sigma+1])
        xx_i        = xx[ind_i]
        yy_i        = sigma(xx_i,x_bp[i_sigma],x_bp[i_sigma+1],y0=y_bp[i_sigma],y1=y_bp[i_sigma+1])
        yy[ind_i]   = yy_i

    # add first and last values to beginning and end
    yy[xx<x_bp[0]]      = y_bp[0]
    yy[xx>=x_bp[-1]]     = y_bp[-1]

    if False:  # debug plot
        import matplotlib.pyplot as plt
        plt.plot(xx,yy)
        plt.show()

    return yy



from numpy import zeros, array
import numpy as np
from ROSCO_toolbox.ofTools.fast_io.FAST_vars_out import FstOutput, Fst7Output

# This variable tree contains all parameters required to create a FAST model
# for FAST versions 7 and 8.


# .fst Simulation Control
Fst = {}
Fst['Echo']            = False         
Fst['AbortLevel']      = ''            
Fst['TMax']            = 0.0           
Fst['DT']              = 0.0           
Fst['InterpOrder']     = 0             
Fst['NumCrctn']        = 0             
Fst['DT_UJac']         = 0.0           
Fst['UJacSclFact']     = 0.0           

# Feature Switches and Flags
Fst['CompElast']    = 0             
Fst['CompInflow']   = 0             
Fst['CompAero']     = 0             
Fst['CompServo']    = 0             
Fst['CompHydro']    = 0             
Fst['CompSub']      = 0             
Fst['CompMooring']  = 0             
Fst['CompIce']      = 0             
Fst['CompNoise']    = 0             #FAST7 only

# Input Files
Fst['EDFile']          = ''            
Fst['BDBldFile1']      = ''            
Fst['BDBldFile2']      = ''            
Fst['BDBldFile3']      = ''            
Fst['InflowFile']      = ''            
Fst['AeroFile']        = ''            
Fst['ServoFile']       = ''            
Fst['HydroFile']       = ''            
Fst['SubFile']         = ''            
Fst['MooringFile']     = ''            
Fst['IceFile']         = ''            

# FAST Output Parameters
Fst['SumPrint']   = False         
Fst['SttsTime']   = 0.0           
Fst['ChkptTime']  = 0.0           
Fst['DT_Out']     = 0.0           
Fst['TStart']     = 0.0           
Fst['OutFileFmt'] = 0             
Fst['TabDelim']   = False         
Fst['OutFmt']     = ''            

# Fst
Fst['Linearize']    = False         
Fst['NLinTimes']    = 2             
Fst['LinTimes']     = (30,          60)
Fst['LinInputs']    = 1             
Fst['LinOutputs']   = 1             
Fst['LinOutJac']    = False         
Fst['LinOutMod']    = False         

# Fst
Fst['WrVTK']        = 0             
Fst['VTK_type']     = 0             
Fst['VTK_fields']   = False         
Fst['VTK_fps']      = 0             

# ElastoDyn Simulation Control
ElastoDyn = {}
ElastoDyn['Echo']             = False         
ElastoDyn['Method']           = 0             
ElastoDyn['DT']               = 0.0           

# Environmental Condition
ElastoDyn['Gravity']          = 0.0           

# Degrees of Freedom
ElastoDyn['FlapElastoDyn1']               = False         
ElastoDyn['FlapElastoDyn2']               = False         
ElastoDyn['EdgeElastoDyn']                = False         
ElastoDyn['TeetElastoDyn']                = False         
ElastoDyn['DrTrElastoDyn']                = False         
ElastoDyn['GenElastoDyn']                 = False         
ElastoDyn['YawElastoDyn']                 = False         
ElastoDyn['TwFAElastoDyn1']               = False         
ElastoDyn['TwFAElastoDyn2']               = False         
ElastoDyn['TwSSElastoDyn1']               = False         
ElastoDyn['TwSSElastoDyn2']               = False         
ElastoDyn['PtfmSgElastoDyn']              = False         
ElastoDyn['PtfmSwElastoDyn']              = False         
ElastoDyn['PtfmHvElastoDyn']              = False         
ElastoDyn['PtfmRElastoDyn']               = False         
ElastoDyn['PtfmPElastoDyn']               = False         
ElastoDyn['PtfmYElastoDyn']               = False         

# Initial Conditions
ElastoDyn['OoPDefl']          = 0.0           
ElastoDyn['IPDefl']           = 0.0           
ElastoDyn['BlPitch1']         = 0.0           
ElastoDyn['BlPitch2']         = 0.0           
ElastoDyn['BlPitch3']         = 0.0           
ElastoDyn['TeetDefl']         = 0.0           
ElastoDyn['Azimuth']          = 0.0           
ElastoDyn['RotSpeed']         = 0.0           
ElastoDyn['NacYaw']           = 0.0           
ElastoDyn['TTDspFA']          = 0.0           
ElastoDyn['TTDspSS']          = 0.0           
ElastoDyn['PtfmSurge']        = 0.0           
ElastoDyn['PtfmSway']         = 0.0           
ElastoDyn['PtfmHeave']        = 0.0           
ElastoDyn['PtfmRoll']         = 0.0           
ElastoDyn['PtfmPitch']        = 0.0           
ElastoDyn['PtfmYaw']          = 0.0           

# Turbine Configuration
ElastoDyn['NumBl']           = 0             
ElastoDyn['TipRad']          = 0.0           
ElastoDyn['HubRad']          = 0.0           
ElastoDyn['PreCone(1)']        = 0.0           
ElastoDyn['PreCone(2)']        = 0.0           
ElastoDyn['PreCone(3)']        = 0.0           
ElastoDyn['HubCM']           = 0.0           
ElastoDyn['UndSling']        = 0.0           
ElastoDyn['Delta3']          = 0.0           
ElastoDyn['AzimB1Up']        = 0.0           
ElastoDyn['OverHang']        = 0.0           
ElastoDyn['ShftGagL']        = 0.0           
ElastoDyn['ShftTilt']        = 0.0           
ElastoDyn['NacCMxn']         = 0.0           
ElastoDyn['NacCMyn']         = 0.0           
ElastoDyn['NacCMzn']         = 0.0           
ElastoDyn['NcIMUxn']         = 0.0           
ElastoDyn['NcIMUyn']         = 0.0           
ElastoDyn['NcIMUzn']         = 0.0           
ElastoDyn['Twr2Shft']        = 0.0           
ElastoDyn['TowerHt']         = 0.0           
ElastoDyn['TowerBsHt']       = 0.0           
ElastoDyn['PtfmCMxt']        = 0.0           
ElastoDyn['PtfmCMyt']        = 0.0           
ElastoDyn['PtfmCMzt']        = 0.0           
ElastoDyn['PtfmRefzt']       = 0.0           

# Mass and Inertia
ElastoDyn['TipMass(1)']       = 0.0           
ElastoDyn['TipMass(2)']       = 0.0           
ElastoDyn['TipMass(3)']       = 0.0           
ElastoDyn['HubMass']        = 0.0           
ElastoDyn['HubIner']        = 0.0           
ElastoDyn['GenIner']        = 0.0           
ElastoDyn['NacMass']        = 0.0           
ElastoDyn['NacYIner']       = 0.0           
ElastoDyn['YawBrMass']      = 0.0           
ElastoDyn['PtfmMass']       = 0.0           
ElastoDyn['PtfmRIner']      = 0.0           
ElastoDyn['PtfmPIner']      = 0.0           
ElastoDyn['PtfmYIner']      = 0.0           

# ED Blade (Structure)
ElastoDyn['BldNodes']        = 0             
ElastoDyn['BldFile1']        = ''            
ElastoDyn['BldFile2']        = ''            
ElastoDyn['BldFile3']        = ''            

# Including the blade files and properties in the same object,
# as is done here, implies that the properties are done for all
# blades (assumed for now)

# General Model Inputs
ElastoDynBlade = {}
ElastoDynBlade['NBlInpSt']        = 0             #Number of blade input stations (-)
ElastoDynBlade['BldFlDmp1']       = 0.0           #Blade flap mode #1 structural damping in percent of critical (%)
ElastoDynBlade['BldFlDmp2']       = 0.0           #Blade flap mode #2 structural damping in percent of critical (%)
ElastoDynBlade['BldEdDmp1']       = 0.0           #Blade edge mode #1 structural damping in percent of critical (%)
ElastoDynBlade['FlStTunr1']       = 0.0           #Blade flapwise modal stiffness tuner, 1st mode (-)
ElastoDynBlade['FlStTunr2']       = 0.0           #Blade flapwise modal stiffness tuner, 2nd mode (-)
ElastoDynBlade['AdjBlMs']         = 0.0           #Factor to adjust blade mass density (-)
ElastoDynBlade['AdjFlSt']         = 0.0           #Factor to adjust blade flap stiffness (-)
ElastoDynBlade['AdjEdSt']         = 0.0           #Factor to adjust blade edge stiffness (-)
        
# Distributed Blade Properties
ElastoDynBlade['BlFract']         = zeros([1])    
ElastoDynBlade['AeroCent']        = zeros([1])    
ElastoDynBlade['PitchAxis']       = zeros([1])    
ElastoDynBlade['StrcTwst']        = zeros([1])    
ElastoDynBlade['BMassDen']        = zeros([1])    
ElastoDynBlade['FlpStff']         = zeros([1])    
ElastoDynBlade['EdgStff']         = zeros([1])    
ElastoDynBlade['GJStff']          = zeros([1])    
ElastoDynBlade['EAStff']          = zeros([1])    
ElastoDynBlade['Alpha']           = zeros([1])    
ElastoDynBlade['FlpIner']         = zeros([1])    
ElastoDynBlade['EdgIner']         = zeros([1])    
ElastoDynBlade['PrecrvRef']       = zeros([1])    
ElastoDynBlade['PreswpRef']       = zeros([1])    #[AH] Added during openmdao1 update
ElastoDynBlade['FlpcgOf']         = zeros([1])    
ElastoDynBlade['Edgcgof']         = zeros([1])    
ElastoDynBlade['FlpEAOf']         = zeros([1])    
ElastoDynBlade['EdgEAOf']         = zeros([1])    
        
# Blade Mode Shapes
ElastoDynBlade['BldFl1Sh']        = zeros([1])    
ElastoDynBlade['BldFl2Sh']        = zeros([1])    
ElastoDynBlade['BldEdgSh']        = zeros([1])    

# Rotor-Teeter
ElastoDyn['TeetMod']        = 0             
ElastoDyn['TeetDmpP']       = 0.0           
ElastoDyn['TeetDmp']        = 0.0           
ElastoDyn['TeetCDmp']       = 0.0           
ElastoDyn['TeetSStP']       = 0.0           
ElastoDyn['TeetHStP']       = 0.0           
ElastoDyn['TeetSSSp']       = 0.0           
ElastoDyn['TeetHSSp']       = 0.0           

ElastoDyn['GBoxEff']         = 0.0           
ElastoDyn['GBRatio']         = 0.0           
ElastoDyn['DTTorSpr']        = 0.0           
ElastoDyn['DTTorDmp']        = 0.0           

ElastoDyn['ElastoDyn']            = False         
ElastoDyn['FurlFile']           = ''            

ElastoDyn['TwrNodes']             = 0             
ElastoDyn['TwrFile']              = ''            

# General Tower Parameters
ElastoDynTower = {}
ElastoDynTower['NTwInptSt']            = 0             #Number of input stations to specify tower geometry
ElastoDynTower['CalcTMode']            = False         #calculate tower mode shapes internally {T: ignore mode shapes from below, F: use mode shapes from below} [CURRENTLY IGNORED] (flag)
ElastoDynTower['TwrFADmp1']            = 0.0           #Tower 1st fore-aft mode structural damping ratio (%)
ElastoDynTower['TwrFADmp2']            = 0.0           #Tower 2nd fore-aft mode structural damping ratio (%)
ElastoDynTower['TwrSSDmp1']            = 0.0           #Tower 1st side-to-side mode structural damping ratio (%)
ElastoDynTower['TwrSSDmp2']            = 0.0           #Tower 2nd side-to-side mode structural damping ratio (%)

# Tower Adjustment Factors
ElastoDynTower['FAStTunr1']            = 0.0           #Tower fore-aft modal stiffness tuner, 1st mode (-)
ElastoDynTower['FAStTunr2']            = 0.0           #Tower fore-aft modal stiffness tuner, 2nd mode (-)
ElastoDynTower['SSStTunr1']            = 0.0           #Tower side-to-side stiffness tuner, 1st mode (-)
ElastoDynTower['SSStTunr2']            = 0.0           #Tower side-to-side stiffness tuner, 2nd mode (-)
ElastoDynTower['AdjTwMa']              = 0.0           #Factor to adjust tower mass density (-)
ElastoDynTower['AdjFASt']              = 0.0           #Factor to adjust tower fore-aft stiffness (-)
ElastoDynTower['AdjSSSt']              = 0.0           #Factor to adjust tower side-to-side stiffness (-)
     
# Distributed Tower Properties
ElastoDynTower['HtFract']              = zeros([1])    
ElastoDynTower['TMassDen']             = zeros([1])    
ElastoDynTower['TwFAStif']             = zeros([1])    
ElastoDynTower['TwSSStif']             = zeros([1])    
ElastoDynTower['TwGJStif']             = zeros([1])    
ElastoDynTower['TwEAStif']             = zeros([1])    
ElastoDynTower['TwFAIner']             = zeros([1])    
ElastoDynTower['TwSSIner']             = zeros([1])    
ElastoDynTower['TwFAcgOf']             = zeros([1])    
ElastoDynTower['TwSScgOf']             = zeros([1])    
        
# Tower Mode Shapes
ElastoDynTower['TwFAM1Sh']             = zeros([1])    #Tower Fore-Aft Mode 1 Shape Coefficients x^2, x^3, x^4, x^5, x^6
ElastoDynTower['TwFAM2Sh']             = zeros([1])    #Tower Fore-Aft Mode 2 Shape Coefficients x^2, x^3, x^4, x^5, x^6
ElastoDynTower['TwSSM1Sh']             = zeros([1])    #Tower Side-to-Side Mode 1 Shape Coefficients x^2, x^3, x^4, x^5, x^6
ElastoDynTower['TwSSM2Sh']             = zeros([1])    #Tower Side-to-Side Mode 2 Shape Coefficients x^2, x^3, x^4, x^5, x^6

ElastoDyn = {}
ElastoDyn['SumPrint']       = False         
ElastoDyn['OutFile']        = 0             
ElastoDyn['TabDelim']       = False         
ElastoDyn['OutFmt']         = ''            
ElastoDyn['TStart']         = 0.0           
ElastoDyn['DecFact']        = 0.0           
ElastoDyn['NTwGages']       = 0             
ElastoDyn['TwrGagNd']       = []            
ElastoDyn['NBlGages']       = 0             
ElastoDyn['BldGagNd']       = []            

# Inflow Wind General Parameters
InflowWind = {}
InflowWind['Echo']            = False         
InflowWind['WindType']        = 0             
InflowWind['PropagationDir']  = 0.0           
InflowWind['NWindVel']        = 0             
InflowWind['WindVxiList']     = 0.0           
InflowWind['WindVyiList']     = 0.0           
InflowWind['WindVziList']     = 0.0           

# Parameters for Steady Wind Conditions [used only for WindType = 1]
InflowWind['HWindSpeed'] = 0.0           
InflowWind['RefHt']     = 0.0           
InflowWind['PLexp']     = 0.0           

# Parameters for Uniform wind file [used only for WindType = 2]
InflowWind['UniformFilename'] = ''            
InflowWind['RefHt']    = 0.0           
InflowWind['RefLength'] = 0.0           

# Parameters for Binary TurbSim Full-Field files [used only for WindType = 3]
InflowWind['TurbSimFilename'] = ''            

# Parameters for Binary Bladed-style Full-Field files [used only for WindType = 4]
InflowWind['FilenameRoot'] = ''            
InflowWind['TowerFile'] = False         

# Parameters for HAWC-format binary files [Only used with WindType = 5]
InflowWind['FileName_u']  = ''            
InflowWind['FileName_v']  = ''            
InflowWind['FileName_w']  = ''            
InflowWind['nx']          = 0             
InflowWind['ny']          = 0             
InflowWind['nz']          = 0             
InflowWind['dx']          = 0.0           
InflowWind['dy']          = 0.0           
InflowWind['dz']          = 0.0           
InflowWind['RefHt']       = 0.0           
InflowWind['ScaleMethod'] = 0             
InflowWind['SFx']         = 0.0           
InflowWind['SFy']         = 0.0           
InflowWind['SFz']         = 0.0           
InflowWind['SigmaFx']     = 0.0           
InflowWind['SigmaFy']     = 0.0           
InflowWind['SigmaFz']     = 0.0           
InflowWind['URef']        = 0.0           
InflowWind['WindProfile'] = 0             
InflowWind['PLExp']       = 0.0           
InflowWind['Z0']          = 0.0           
InflowWind['InitPosition(x)'] = 0.0           

# Inflow Wind Output Parameters (actual OutList included in master OutList)
InflowWind['SumPrint']   = False         

# # Wnd Wind File Parameters
# WndWind = {}
# WndWind['TimeSteps']          = 0             #number of time steps
# WndWind['Time']               = zeros([1])    #time steps
# WndWind['HorSpd']             = zeros([1])    #horizontal wind speed
# WndWind['WindDir']            = zeros([1])    #wind direction
# WndWind['VerSpd']             = zeros([1])    #vertical wind speed
# WndWind['HorShr']             = zeros([1])    #horizontal shear
# WndWind['VerShr']             = zeros([1])    #vertical power-law shear
# WndWind['LnVShr']             = zeros([1])    #vertical linear shear
# WndWind['GstSpd']             = zeros([1])    #gust speed not sheared by Aerodyn

# AeroDyn Parameters
AeroDyn14 = {}
# General Model Inputs
AeroDyn14['StallMod']         = ""
AeroDyn14['UseCm']            = ""
AeroDyn14['InfModel']         = ""
AeroDyn14['IndModel']         = ""
AeroDyn14['AToler']           = 0.
AeroDyn14['TLModel']          = ""
AeroDyn14['HLModel']          = ""
AeroDyn14['TwrShad']          = ""
AeroDyn14['TwrPotent']        = False
AeroDyn14['TwrShadow']        = False
AeroDyn14['TwrFile']          = ""
AeroDyn14['CalcTwrAero']      = False
AeroDyn14['AirDens']          = 0.0
AeroDyn14['KinVisc']          = 0.0
AeroDyn14['DTAero']           = "default"
AeroDyn14['NumFoil']          = 0
AeroDyn14['FoilNm']           = [""]

AeroDynBlade = {}
AeroDynBlade['BldNodes']         = 0
AeroDynBlade['RNodes']           = np.asarray([]) 
AeroDynBlade['AeroTwst']         = np.asarray([]) 
AeroDynBlade['DRNodes']          = np.asarray([]) 
AeroDynBlade['Chord']            = np.asarray([]) 
AeroDynBlade['NFoil']            = np.asarray([]) 
AeroDynBlade['PrnElm']           = np.asarray([]) 


# AeroDyn Blade
AeroDynTower = {}
AeroDynTower['NTwrHt']        = 0             
AeroDynTower['NTwrRe']        = 0             
AeroDynTower['NTwrCD']        = 0             
AeroDynTower['Tower_Wake_Constant'] = 0.            
AeroDynTower['TwrHtFr']       = np.asarray([]) 
AeroDynTower['TwrWid']        = np.asarray([]) 
AeroDynTower['NTwrCDCol']     = np.asarray([]) 
AeroDynTower['TwrRe']         = np.asarray([]) 
AeroDynTower['TwrCD']         = np.asarray([]) 


# AeroDyn Airfoil Polar
AeroDynPolar = {}
AeroDynPolar['IDParam']     = 0.0           #Table ID Parameter (Typically Reynolds number)
AeroDynPolar['StallAngle']  = 0.0           #Stall angle (deg)
AeroDynPolar['ZeroCn']      = 0.0           #Zero lift angle of attack (deg)
AeroDynPolar['CnSlope']     = 0.0           #Cn slope for zero lift (dimensionless)
AeroDynPolar['CnPosStall']  = 0.0           #Cn at stall value for positive angle of attack
AeroDynPolar['CnNegStall']  = 0.0           #Cn at stall value for negative angle of attack
AeroDynPolar['alphaCdMin']  = 0.0           #Angle of attack for minimum CD (deg)
AeroDynPolar['CdMin']       = 0.0           #Minimum Cd Value

AeroDynPolar['alpha']       = zeros([1])    #angle of attack
AeroDynPolar['cl']          = zeros([1])    #coefficient of lift
AeroDynPolar['cd']          = zeros([1])    #coefficient of drag
AeroDynPolar['cm']          = zeros([1])    #coefficient of the pitching moment

AeroDynPolar['Re']          = 0.0
AeroDynPolar['Ctrl']        = 0.0
AeroDynPolar['InclUAdata']  = 0.0
AeroDynPolar['alpha0']      = 0.0
AeroDynPolar['alpha1']      = 0.0
AeroDynPolar['alpha2']      = 0.0
AeroDynPolar['eta_e']       = 0.0
AeroDynPolar['C_nalpha']    = 0.0
AeroDynPolar['T_f0']        = 0.0
AeroDynPolar['T_V0']        = 0.0
AeroDynPolar['T_p']         = 0.0
AeroDynPolar['T_VL']        = 0.0
AeroDynPolar['b1']          = 0.0
AeroDynPolar['b2']          = 0.0
AeroDynPolar['b5']          = 0.0
AeroDynPolar['A1']          = 0.0
AeroDynPolar['A2']          = 0.0
AeroDynPolar['A5']          = 0.0
AeroDynPolar['S1']          = 0.0
AeroDynPolar['S2']          = 0.0
AeroDynPolar['S3']          = 0.0
AeroDynPolar['S4']          = 0.0
AeroDynPolar['Cn1']         = 0.0
AeroDynPolar['Cn2']         = 0.0
AeroDynPolar['St_sh']       = 0.0
AeroDynPolar['Cd0']         = 0.0
AeroDynPolar['Cm0']         = 0.0
AeroDynPolar['k0']          = 0.0
AeroDynPolar['k1']          = 0.0
AeroDynPolar['k2']          = 0.0
AeroDynPolar['k3']          = 0.0
AeroDynPolar['k1_hat']      = 0.0
AeroDynPolar['x_cp_bar']    = 0.0
AeroDynPolar['UACutout']    = 0.0
AeroDynPolar['filtCutOff']  = 0.0

# AeroDyn15
AeroDyn15 = {}
AeroDyn15['Echo']           = False
AeroDyn15['DTAero']         = 0.0
AeroDyn15['WakeMod']        = 0
AeroDyn15['AFAeroMod']      = 0
AeroDyn15['TwrPotent']      = 0
AeroDyn15['TwrShadow']      = False
AeroDyn15['TwrAero']        = False
AeroDyn15['FrozenWake']     = False
AeroDyn15['CavitCheck']     = False

AeroDyn15['AirDens']        = 0.0
AeroDyn15['KinVisc']        = 0.0
AeroDyn15['SpdSound']       = 0.0
AeroDyn15['Patm']           = 0.0
AeroDyn15['Pvap']           = 0.0
AeroDyn15['FluidDepth']     = 0.0

AeroDyn15['SkewMod']        = 0
AeroDyn15['SkewModFactor']  = "default"
AeroDyn15['TipLoss']        = False
AeroDyn15['HubLoss']        = False
AeroDyn15['TanInd']         = False
AeroDyn15['AIDrag']         = False
AeroDyn15['TIDrag']         = False
AeroDyn15['IndToler']       = 0.0
AeroDyn15['MaxIter']        = 0

AeroDyn15['DBEMT_Mod']      = 2
AeroDyn15['tau1_const']     = 4

AeroDyn15['UAMod']          = 0
AeroDyn15['FLookup']        = False

AeroDyn15['InCol_Alfa']        = 0
AeroDyn15['InCol_Cl']          = 0
AeroDyn15['InCol_Cd']          = 0
AeroDyn15['InCol_Cm']          = 0
AeroDyn15['InCol_Cpmin']       = 0
AeroDyn15['NumAFfiles']        = 0
AeroDyn15['AFNames']           = []

AeroDyn15['UseBlCm']           = False
AeroDyn15['ADBlFile1']         = ''
AeroDyn15['ADBlFile2']         = ''
AeroDyn15['ADBlFile3']         = ''

AeroDyn15['NumTwrNds']         = 0
AeroDyn15['TwrElev']           = []
AeroDyn15['TwrDiam']           = []
AeroDyn15['TwrCd']             = []
AeroDyn15['TwrElev']           = [] 
AeroDyn15['TwrDiam']           = [] 
AeroDyn15['TwrCd']             = []

AeroDyn15['SumPrint']          = False
AeroDyn15['NBlOuts']           = 0
AeroDyn15['BlOutNd']           = []
AeroDyn15['NTwOuts']           = 0
AeroDyn15['TwOutNd']           = []


# ServoDyn Simulation Control
ServoDyn = {}
ServoDyn['Echo']             = False         
ServoDyn['DT']               = 0.0           

# Pitch Control
ServoDyn['PCMode']           = 0             
ServoDyn['TPCOn']            = 0.0           
ServoDyn['TPitManS1']        = 0.0           
ServoDyn['TPitManS2']        = 0.0           
ServoDyn['TPitManS3']        = 0.0           
ServoDyn['TPitManE1']        = 0.0           #FAST7 only
ServoDyn['TPitManE2']        = 0.0           #FAST7 only
ServoDyn['TPitManE3']        = 0.0           #FAST7 only
ServoDyn['PitManRat(1)']     = 0.0           
ServoDyn['PitManRat(2)']     = 0.0           
ServoDyn['PitManRat(3)']     = 0.0           
ServoDyn['BlPitchF(1)']      = 0.0           
ServoDyn['BlPitchF(2)']      = 0.0           
ServoDyn['BlPitchF(3)']      = 0.0           
ServoDyn['BlPitch1']         = 0.0           #FAST7 only
ServoDyn['BlPitch2']         = 0.0           #FAST7 only
ServoDyn['BlPitch3']         = 0.0           #FAST7 only

# Generator and Torque Control
ServoDyn['VSContrl']         = 0             
ServoDyn['GenModel']         = 0             
ServoDyn['GenEff']           = 0.0           
ServoDyn['GenTiStr']         = False         
ServoDyn['GenTiStp']         = False         
ServoDyn['SpdGenOn']         = 0.0           
ServoDyn['TimGenOn']         = 0.0           
ServoDyn['TimGenOf']         = 0.0           

# Simple Variable-Speed Torque Control
ServoDyn['VS_RtGnSp']        = 0.0           
ServoDyn['VS_RtTq']          = 0.0           
ServoDyn['VS_Rgn2K']         = 0.0           
ServoDyn['VS_SlPc']          = 0.0           

# Simple Induction Generator
ServoDyn['SIG_SlPc']         = 0.0           
ServoDyn['SIG_SySp']         = 0.0           
ServoDyn['SIG_RtTq']         = 0.0           
ServoDyn['SIG_PORt']         = 0.0           

# Thevenin-Equivalent Induction Generator
ServoDyn['TEC_Freq']         = 0.0           
ServoDyn['TEC_NPol']         = 0             
ServoDyn['TEC_SRes']         = 0.0           
ServoDyn['TEC_RRes']         = 0.0           
ServoDyn['TEC_VLL']          = 0.0           
ServoDyn['TEC_SLR']          = 0.0           
ServoDyn['TEC_RLR']          = 0.0           
ServoDyn['TEC_MR']           = 0.0           

# High-Speed Shaft Brake
ServoDyn['HSSBrMode']        = 0             
ServoDyn['THSSBrDp']         = 0.0           
ServoDyn['HSSBrDT']          = 0.0           
ServoDyn['HSSBrTqF']         = 0.0           

# Nacelle-Yaw Control
ServoDyn['YCMode']           = 0             
ServoDyn['TYCOn']            = 0.0           
ServoDyn['YawNeut']          = 0.0           
ServoDyn['YawSpr']           = 0.0           
ServoDyn['YawDamp']          = 0.0           
ServoDyn['TYawManS']         = 0.0           
ServoDyn['YawManRat']        = 0.0           
ServoDyn['NacYawF']          = 0.0           

# Tip Brake (used in FAST7 only)
ServoDyn['TiDynBrk']         = 0.0           
ServoDyn['TTpBrDp1']         = 0.0           
ServoDyn['TTpBrDp2']         = 0.0           
ServoDyn['TTpBrDp3']         = 0.0           
ServoDyn['TBDepISp1']        = 0.0           
ServoDyn['TBDepISp2']        = 0.0           
ServoDyn['TBDepISp3']        = 0.0           
ServoDyn['TBDrConN']         = 0.0           
ServoDyn['TBDrConD']         = 0.0           
ServoDyn['TpBrDT']           = 0.0           

# Tuned Mass Damper
ServoDyn = {}
ServoDyn['CompNTMD']         = False         
ServoDyn['NTMDfile']         = ''            
ServoDyn['CompTTMD']         = False         
ServoDyn['TTMDfile']         = ''            

# Bladed Interface
ServoDyn['DLL_FileName']     = ''            
ServoDyn['DLL_InFile']       = ''            
ServoDyn['DLL_ProcName']     = ''            
ServoDyn['DLL_DT']           = ''            
ServoDyn['DLL_Ramp']         = False         
ServoDyn['BPCutoff']         = 0.0           
ServoDyn['NacYaw_North']     = 0.0           
ServoDyn['Ptch_Cntrl']       = 0.0           
ServoDyn['Ptch_SetPnt']      = 0.0           
ServoDyn['Ptch_Min']         = 0.0           
ServoDyn['Ptch_Max']         = 0.0           
ServoDyn['PtchRate_Min']     = 0.0           
ServoDyn['PtchRate_Max']     = 0.0           
ServoDyn['Gain_OM']          = 0.0           
ServoDyn['GenSpd_MinOM']     = 0.0           
ServoDyn['GenSpd_MaxOM']     = 0.0           
ServoDyn['GenSpd_Dem']       = 0.0           
ServoDyn['GenTrq_Dem']       = 0.0           
ServoDyn['GenPwr_Dem']       = 0.0           
ServoDyn['DLL_NumTrq']       = 0.0           
ServoDyn['GenSpd_TLU']       = zeros([0])    
ServoDyn['GenTrq_TLU']       = zeros([0])    

# ServoDyn Output Params
ServoDyn['SumPrint']         = False         
ServoDyn['OutFile']          = 0             
ServoDyn['TabDelim']         = False         
ServoDyn['OutFmt']           = ''            
ServoDyn['TStart']           = 0.0           

# Bladed style Interface controller input file, intended for ROSCO https://github.com/NREL/ROSCO_toolbox
DISCON_in = {}
DISCON_in['LoggingLevel']      = 0
DISCON_in['F_LPFType']         = 0
DISCON_in['F_NotchType']       = 0
DISCON_in['IPC_ControlMode']   = 0
DISCON_in['VS_ControlMode']    = 0
DISCON_in['PC_ControlMode']    = 0
DISCON_in['Y_ControlMode']     = 0
DISCON_in['SS_Mode']           = 0
DISCON_in['WE_Mode']           = 0
DISCON_in['PS_Mode']           = 0

DISCON_in['F_LPFCornerFreq']   = 0.0
DISCON_in['F_LPFDamping']      = 0.0
DISCON_in['F_NotchCornerFreq'] = 0.0
DISCON_in['F_NotchBetaNumDen'] = []
DISCON_in['F_SSCornerFreq']    = 0.0

DISCON_in['PC_GS_n']           = 0
DISCON_in['PC_GS_angles']      = []
DISCON_in['PC_GS_KP']          = []
DISCON_in['PC_GS_KI']          = []
DISCON_in['PC_GS_KD']          = []
DISCON_in['PC_GS_TF']          = []
DISCON_in['PC_MaxPit']         = 0.0
DISCON_in['PC_MinPit']         = 0.0
DISCON_in['PC_MaxRat']         = 0.0
DISCON_in['PC_MinRat']         = 0.0
DISCON_in['PC_RefSpd']         = 0.0
DISCON_in['PC_FinePit']        = 0.0
DISCON_in['PC_Switch']         = 0.0
DISCON_in['Z_EnableSine']      = 0
DISCON_in['Z_PitchAmplitude']  = 0.0
DISCON_in['Z_PitchFrequency']  = 0.0

DISCON_in['IPC_IntSat']        = 0.0
DISCON_in['IPC_KI']            = []
DISCON_in['IPC_aziOffset']     = []
DISCON_in['IPC_CornerFreqAct'] = 0.0

DISCON_in['VS_GenEff']         = 0.0
DISCON_in['VS_ArSatTq']        = 0.0
DISCON_in['VS_MaxRat']         = 0.0
DISCON_in['VS_MaxTq']          = 0.0
DISCON_in['VS_MinTq']          = 0.0
DISCON_in['VS_MinOMSpd']       = 0.0
DISCON_in['VS_Rgn2K']          = 0.0
DISCON_in['VS_RtPwr']          = 0.0
DISCON_in['VS_RtTq']           = 0.0
DISCON_in['VS_RefSpd']         = 0.0
DISCON_in['VS_n']              = 0
DISCON_in['VS_KP']             = 0.0
DISCON_in['VS_KI']             = 0.0
DISCON_in['VS_TSRopt']         = 0.0

DISCON_in['SS_VSGain']         = 0.0
DISCON_in['SS_PCGain']         = 0.0

DISCON_in['WE_BladeRadius']    = 0.0
DISCON_in['WE_CP_n']           = 0
DISCON_in['WE_CP']             = []
DISCON_in['WE_Gamma']          = 0.0
DISCON_in['WE_GearboxRatio']   = 0.0
DISCON_in['WE_Jtot']           = 0.0
DISCON_in['WE_RhoAir']         = 0.0
DISCON_in['PerfFileName']      = ""
DISCON_in['PerfTableSize']     = []
DISCON_in['WE_FOPoles_N']      = 0
DISCON_in['WE_FOPoles_v']      = []
DISCON_in['WE_FOPoles']        = []

DISCON_in['Y_ErrThresh']       = 0.0
DISCON_in['Y_IPC_IntSat']      = 0.0
DISCON_in['Y_IPC_n']           = 0
DISCON_in['Y_IPC_KP']          = 0.0
DISCON_in['Y_IPC_KI']          = 0.0
DISCON_in['Y_IPC_omegaLP']     = 0.0
DISCON_in['Y_IPC_zetaLP']      = 0.0
DISCON_in['Y_MErrSet']         = 0.0
DISCON_in['Y_omegaLPFast']     = 0.0
DISCON_in['Y_omegaLPSlow']     = 0.0
DISCON_in['Y_Rate']            = 0.0

DISCON_in['FA_KI']             = 0.0
DISCON_in['FA_HPF_CornerFreq'] = 0.0
DISCON_in['FA_IntSat']         = 0.0
DISCON_in['PS_BldPitchMin_N']  = 0
DISCON_in['PS_WindSpeeds']     = []
DISCON_in['PS_BldPitchMin']    = []

# HydroDyn Input File
HydroDyn = {}
HydroDyn['Echo']             = False

# ENVIRONMENTAL CONDITIONS
HydroDyn['WtrDens']          = 0.
HydroDyn['WtrDpth']          = 0.
HydroDyn['MSL2SWL']          = 0.

# WAVES
HydroDyn['WaveMod']          = 0
HydroDyn['WaveStMod']        = 0
HydroDyn['WaveTMax']         = 0.
HydroDyn['WaveDT']           = 0.
HydroDyn['WaveHs']           = 0.
HydroDyn['WaveTp']           = 0.
HydroDyn['WavePkShp']        = "DEFAULT"
HydroDyn['WvLowCOff']        = 0.
HydroDyn['WvHiCOff']         = 0.
HydroDyn['WaveDir']          = 0.
HydroDyn['WaveDirMod']       = 0
HydroDyn['WaveDirSpread']    = 0.
HydroDyn['WaveNDir']         = 0
HydroDyn['WaveDirRange']     = 0.
HydroDyn['WaveSeed1']        = 0.
HydroDyn['WaveSeed2']        = 0.
HydroDyn['WaveNDAmp']        = False
HydroDyn['WvKinFile']        = ""
HydroDyn['NWaveElev']        = 0.
HydroDyn['WaveElevxi']       = 0.
HydroDyn['WaveElevyi']       = 0.

# 2ND-ORDER WAVES
HydroDyn['WvDiffQTF']        = False
HydroDyn['WvSumQTF']         = False
HydroDyn['WvLowCOffD']       = 0.
HydroDyn['WvHiCOffD']        = 0.
HydroDyn['WvLowCOffS']       = 0.
HydroDyn['WvHiCOffS']        = 0.

# CURRENT
HydroDyn['CurrMod']          = 0
HydroDyn['CurrSSV0']         = 0.
HydroDyn['CurrSSDir']        = "DEFAULT"
HydroDyn['CurrNSRef']        = 0.
HydroDyn['CurrNSV0']         = 0.
HydroDyn['CurrNSDir']        = 0.
HydroDyn['CurrDIV']          = 0.
HydroDyn['CurrDIDir']        = 0.

# FLOATING PLATFORM
HydroDyn['PotMod']           = 0
HydroDyn['PotFile']          = ""
HydroDyn['WAMITULEN']        = 0.
HydroDyn['PtfmVol0']         = 0.
HydroDyn['PtfmCOBxt']        = 0.
HydroDyn['PtfmCOByt']        = 0.
HydroDyn['RdtnMod']          = 0
HydroDyn['RdtnTMax']         = 0.
HydroDyn['RdtnDT']           = 0.

# 2ND-ORDER FLOATING PLATFORM FORCES
HydroDyn['MnDrift']          = 0.
HydroDyn['NewmanApp']        = 0.
HydroDyn['DiffQTF']          = 0.
HydroDyn['SumQTF']           = 0.

# FLOATING PLATFORM FORCE FLAGS
HydroDyn['PtfmSgF']          = True
HydroDyn['PtfmSwF']          = True
HydroDyn['PtfmHvF']          = True
HydroDyn['PtfmRF']           = True
HydroDyn['PtfmPF']           = True
HydroDyn['PtfmYF']           = True

# PLATFORM ADDITIONAL STIFFNESS AND DAMPING
HydroDyn['AddF0']            = np.zeros((1,6))
HydroDyn['AddCLin']          = np.zeros((6,6))
HydroDyn['AddBLin']          = np.zeros((6,6))
HydroDyn['AddBQuad']         = np.zeros((6,6))

# AXIAL COEFFICIENTS
HydroDyn['NAxCoef']          = 0
HydroDyn['AxCoefID']         = 0
HydroDyn['AxCd']             = 0.
HydroDyn['AxCa']             = 0.
HydroDyn['AxCp']             = 0.

# MEMBER JOINTS
HydroDyn['NJoints']          = 0
HydroDyn['JointID']          = []
HydroDyn['Jointxi']          = []
HydroDyn['Jointyi']          = []
HydroDyn['Jointzi']          = []
HydroDyn['JointAxID']        = []
HydroDyn['JointOvrlp']       = []

# MEMBER CROSS-SECTION PROPERTIES
HydroDyn['NPropSets']        = 0
HydroDyn['PropSetID']        = []
HydroDyn['PropD']            = []
HydroDyn['PropThck']         = []

# SIMPLE HYDRODYNAMIC COEFFICIENTS
HydroDyn['SimplCd']          = 0.
HydroDyn['SimplCdMG']        = 0.
HydroDyn['SimplCa']          = 0.
HydroDyn['SimplCaMG']        = 0.
HydroDyn['SimplCp']          = 0.
HydroDyn['SimplCpMG']        = 0.
HydroDyn['SimplAxCa']        = 0.
HydroDyn['SimplAxCaMG']      = 0.
HydroDyn['SimplAxCp']        = 0.
HydroDyn['SimplAxCpMG']      = 0.

# DEPTH-BASED HYDRODYNAMIC COEFFICIENTS
HydroDyn['NCoefDpth']        = 0
HydroDyn['Dpth']             = []
HydroDyn['DpthCd']           = []
HydroDyn['DpthCdMG']         = []
HydroDyn['DpthCa']           = []
HydroDyn['DpthCaMG']         = []
HydroDyn['DpthCp']           = []
HydroDyn['DpthCpMG']         = []
HydroDyn['DpthAxCa']         = []
HydroDyn['DpthAxCaMG']       = []
HydroDyn['DpthAxCp']         = []
HydroDyn['DpthAxCpMG']       = []

# MEMBER-BASED HYDRODYNAMIC COEFFICIENTS
HydroDyn['NCoefMembers']     = 0
HydroDyn['MemberID_HydC']    = []
HydroDyn['MemberCd1']        = []
HydroDyn['MemberCd2']        = []
HydroDyn['MemberCdMG1']      = []
HydroDyn['MemberCdMG2']      = []
HydroDyn['MemberCa1']        = []
HydroDyn['MemberCa2']        = []
HydroDyn['MemberCaMG1']      = []
HydroDyn['MemberCaMG2']      = []
HydroDyn['MemberCp1']        = []
HydroDyn['MemberCp2']        = []
HydroDyn['MemberCpMG1']      = []
HydroDyn['MemberCpMG2']      = []
HydroDyn['MemberAxCa1']      = []
HydroDyn['MemberAxCa2']      = []
HydroDyn['MemberAxCaMG1']    = []
HydroDyn['MemberAxCaMG2']    = []
HydroDyn['MemberAxCp1']      = []
HydroDyn['MemberAxCp2']      = []
HydroDyn['MemberAxCpMG1']    = []
HydroDyn['MemberAxCpMG2']    = []

# MEMBERS
HydroDyn['NMembers']         = 0
HydroDyn['MemberID']         = []
HydroDyn['MJointID1']        = []
HydroDyn['MJointID2']        = []
HydroDyn['MPropSetID1']      = []
HydroDyn['MPropSetID2']      = []
HydroDyn['MDivSize']         = []
HydroDyn['MCoefMod']         = []
HydroDyn['PropPot']          = []

# FILLED MEMBERS
HydroDyn['NFillGroups']      = 0
HydroDyn['FillNumM']         = []
HydroDyn['FillMList']        = []
HydroDyn['FillFSLoc']        = []
HydroDyn['FillDens']         = []

# MARINE GROWTH
HydroDyn['NMGDepths']        = 0
HydroDyn['MGDpth']           = []
HydroDyn['MGThck']           = []
HydroDyn['MGDens']           = []

# MEMBER OUTPUT LIST
HydroDyn['NMOutputs']        = 0
HydroDyn['MemberID_out']     = []
HydroDyn['NOutLoc']          = []
HydroDyn['NodeLocs']         = []

# JOINT OUTPUT LIST
HydroDyn['NJOutputs']        = 0
HydroDyn['JOutLst']          = 0

# OUTPUT
HydroDyn['HDSum']            = True
HydroDyn['OutAll']           = False
HydroDyn['OutSwtch']         = 2
HydroDyn['OutFmt']           = ""
HydroDyn['OutSFmt']          = ""

## SubDyn Input File
SubDyn = {}
# SIMULATION CONTROL
SubDyn['Echo']         = False
SubDyn['SDdeltaT']     = 0.
SubDyn['IntMethod']    = 0
SubDyn['SttcSolve']    = False
# FEA and CRAIG-BAMPTON PARAMETERS
SubDyn['FEMMod']       = 0
SubDyn['NDiv']         = 0
SubDyn['CBMod']        = False
SubDyn['Nmodes']       = 0
SubDyn['JDampings']    = 0
# STRUCTURE JOINTS
SubDyn['NJoints']      = 0
SubDyn['JointID']      = [0]
SubDyn['JointXss']     = [0.]
SubDyn['JointYss']     = [0.]
SubDyn['JointZss']     = [0.]
# BASE REACTION JOINTS
SubDyn['NReact']       = 0
SubDyn['RJointID']     = [0]
SubDyn['RctTDXss']     = [0]
SubDyn['RctTDYss']     = [0]
SubDyn['RctTDZss']     = [0]
SubDyn['RctRDXss']     = [0]
SubDyn['RctRDYss']     = [0]
SubDyn['RctRDZss']     = [0]
SubDyn['Rct_SoilFile'] = [""]
# INTERFACE JOINTS
SubDyn['NInterf']      = 0
SubDyn['IJointID']     = [0]
SubDyn['ItfTDXss']     = [0]
SubDyn['ItfTDYss']     = [0]
SubDyn['ItfTDZss']     = [0]
SubDyn['ItfRDXss']     = [0]
SubDyn['ItfRDYss']     = [0]
SubDyn['ItfRDZss']     = [0]
# MEMBERS
SubDyn['NMembers']     = 0
SubDyn['MemberID']     = [0]
SubDyn['MJointID1']    = [0]
SubDyn['MJointID2']    = [0]
SubDyn['MPropSetID1']  = [0]
SubDyn['MPropSetID2']  = [0]
SubDyn['COSMID']       = [0]
# MEMBER X-SECTION PROPERTY data 1/2
SubDyn['NPropSets']    = 0
SubDyn['PropSetID']    = [0.]
SubDyn['YoungE']       = [0.]
SubDyn['ShearG']       = [0.]
SubDyn['MatDens']      = [0.]
SubDyn['XsecD']        = [0.]
SubDyn['XsecT']        = [0.]
# MEMBER X-SECTION PROPERTY data 2/2
SubDyn['NXPropSets']   = 0
SubDyn['PropSetID']    = [0]
SubDyn['YoungE']       = [0.]
SubDyn['ShearG']       = [0.]
SubDyn['MatDens']      = [0.]
SubDyn['XsecA']        = [0.]
SubDyn['XsecAsx']      = [0.]
SubDyn['XsecAsy']      = [0.]
SubDyn['XsecJxx']      = [0.]
SubDyn['XsecJyy']      = [0.]
SubDyn['XsecJ0']       = [0.]
# MEMBER COSINE MATRICES
SubDyn['NCOSMs']       = 0
SubDyn['COSMID']       = 0
SubDyn['COSM11']       = 0.
SubDyn['COSM12']       = 0.
SubDyn['COSM13']       = 0.
SubDyn['COSM21']       = 0.
SubDyn['COSM22']       = 0.
SubDyn['COSM23']       = 0.
SubDyn['COSM31']       = 0.
SubDyn['COSM32']       = 0.
SubDyn['COSM33']       = 0.
# JOINT ADDITIONAL CONCENTRATED MASSES
SubDyn['NCmass']       = 0
SubDyn['CMJointID']    = 0
SubDyn['JMass']        = 0.
SubDyn['JMXX']         = 0.
SubDyn['JMYY']         = 0.
SubDyn['JMZZ']         = 0.
# OUTPUT
SubDyn['SSSum']        = False
SubDyn['OutCOSM']      = False
SubDyn['OutAll']       = False
SubDyn['OutSwtch']     = 0
SubDyn['TabDelim']     = False
SubDyn['OutDec']       = 0
SubDyn['OutFmt']       = ""
SubDyn['OutSFmt']      = ""
# MEMBER OUTPUT LIST
SubDyn['NMOutputs']    = 0
SubDyn['MemberID']     = 0
SubDyn['NOutCnt']      = 0
SubDyn['NodeCnt']      = 0

## MAP++ Input File
# LINE DICTIONARY
MAP = {}
MAP['LineType']              = ""
MAP['Diam']                  = 0.
MAP['MassDenInAir']          = 0.
MAP['EA']                    = 0.
MAP['CB']                    = 0.
MAP['CIntDamp']              = 0.
MAP['Ca']                    = 0.
MAP['Cdn']                   = 0.
MAP['Cdt']                   = 0.

# NODE PROPERTIES
MAP['Node']                  = []
MAP['Type']                  = []
MAP['X']                     = []
MAP['Y']                     = []
MAP['Z']                     = []
MAP['M']                     = []
MAP['B']                     = []
MAP['FX']                    = []
MAP['FY']                    = []
MAP['FZ']                    = []

# LINE PROPERTIES
MAP['Line']                  = 0
MAP['LineType']              = ""
MAP['UnstrLen']              = 0.
MAP['NodeAnch']              = 0
MAP['NodeFair']              = 0
MAP['Flags']                 = []

# SOLVER OPTIONS
MAP['Option']                = []

## MoorDyn Input File
MoorDyn = {}
MoorDyn['Echo']          = False
# LINE TYPES
MoorDyn['NTypes']        = 0
MoorDyn['LineType']      = []
MoorDyn['Diam']          = []
MoorDyn['MassDen']       = []
MoorDyn['EA']            = []
MoorDyn['BA_zeta']       = []
MoorDyn['Can']           = []
MoorDyn['Cat']           = []
MoorDyn['Cdn']           = []
MoorDyn['Cdt']           = []
# CONNECTION PROPERTIES
MoorDyn['NConnects']     = 0
MoorDyn['Node']          = []
MoorDyn['Type']          = []
MoorDyn['X']             = []
MoorDyn['Y']             = []
MoorDyn['Z']             = []
MoorDyn['M']             = []
MoorDyn['V']             = []
MoorDyn['FX']            = []
MoorDyn['FY']            = []
MoorDyn['FZ']            = []
MoorDyn['CdA']           = []
MoorDyn['CA']            = []
# LINE PROPERTIES
MoorDyn['NConnects']     = 0
MoorDyn['Line']          = []
MoorDyn['LineType']      = []
MoorDyn['UnstrLen']      = []
MoorDyn['NumSegs']       = []
MoorDyn['NodeAnch']      = []
MoorDyn['NodeFair']      = []
MoorDyn['Flags_Outputs'] = []
# SOLVER OPTIONS
MoorDyn['dtM']           = 0.
MoorDyn['kbot']          = 0.
MoorDyn['cbot']          = 0.
MoorDyn['dtIC']          = 0.
MoorDyn['TmaxIC']        = 0.
MoorDyn['CdScaleIC']     = 0.
MoorDyn['threshIC']      = 0.


# BeamDyn Input File
# SIMULATION CONTROL
BeamDyn = {}
BeamDyn['Echo']             = False
BeamDyn['QuasiStaticInit']  = True
BeamDyn['rhoinf']           = 0
BeamDyn['quadrature']       = 2
BeamDyn['refine']           = "DEFAULT"
BeamDyn['n_fact']           = "DEFAULT"
BeamDyn['DTBeam']           = "DEFAULT"
BeamDyn['load_retries']     = "DEFAULT"
BeamDyn['NRMax']            = "DEFAULT"
BeamDyn['stop_tol']         = "DEFAULT"
BeamDyn['tngt_stf_fd']      = "DEFAULT"
BeamDyn['tngt_stf_comp']    = "DEFAULT"
BeamDyn['tngt_stf_pert']    = "DEFAULT"
BeamDyn['tngt_stf_difftol'] = "DEFAULT"
BeamDyn['RotStates']        = True
# GEOMETRY PARAMETER
BeamDyn['member_total']     = 1
BeamDyn['kp_total']         = 3
BeamDyn['members']          = [{'kp_xr':[], 'kp_yr':[], 'kp_zr':[], 'initial_twist':[]}]
# MESH PARAMETER
BeamDyn['order_elem']       = 5
# MATERIAL PARAMETER
BeamDyn['BldFile']          = ""
# PITCH ACTUATOR PARAMETERS
BeamDyn['UsePitchAct']      = False
BeamDyn['PitchJ']           = 0.
BeamDyn['PitchK']           = 0.
BeamDyn['PitchC']           = 0.
# OUTPUTS
BeamDyn['SumPrint']         = True
BeamDyn['OutFmt']           = "ES10.3E2"
BeamDyn['NNodeOuts']        = 0.
BeamDyn['OutNd']            = []

## BeamDyn Blade Input File
BeamDynBlade = {}
BeamDynBlade['station_total']   = 1
BeamDynBlade['damp_type']       = 1
BeamDynBlade['mu1']             = 0.01
BeamDynBlade['mu2']             = 0.01
BeamDynBlade['mu3']             = 0.01
BeamDynBlade['mu4']             = 0.01
BeamDynBlade['mu5']             = 0.01
BeamDynBlade['mu6']             = 0.01
BeamDynBlade['radial_stations'] = []
BeamDynBlade['beam_stiff']      = []
BeamDynBlade['beam_inertia']    = []

#######################
Fst7 = {}
Fst7['Echo']        = False
Fst7['ADAMSPrep']   = 0
Fst7['AnalMode']    = 0
Fst7['NumBl']       = 0
Fst7['TMax']        = 0.
Fst7['DT']          = 0.
Fst7['YCMode']      = 0
Fst7['TYCOn']       = 0.
Fst7['PCMode']      = 0
Fst7['TPCOn']       = 0.
Fst7['VSContrl']    = 0
Fst7['VS_RtGnSp']   = 0.
Fst7['VS_RtTq']     = 0.
Fst7['VS_Rgn2K']    = 0.
Fst7['VS_SlPc']     = 0.
Fst7['GenModel']    = 0
Fst7['GenTiStr']    = False
Fst7['GenTiStp']    = False
Fst7['SpdGenOn']    = 0.
Fst7['TimGenOn']    = 0.
Fst7['TimGenOf']    = 0.
Fst7['HSSBrMode']   = 0
Fst7['THSSBrDp']    = 0.
Fst7['TiDynBrk']    = 0.
Fst7['TTpBrDp1']    = 0.
Fst7['TTpBrDp2']    = 0.
Fst7['TTpBrDp3']    = 0.
Fst7['TBDepISp1']   = 0.
Fst7['TBDepISp2']   = 0.
Fst7['TBDepISp3']   = 0.
Fst7['TYawManS']    = 0.
Fst7['TYawManE']    = 0.
Fst7['NacYawF']     = 0.
Fst7['TPitManS1']   = 0.
Fst7['TPitManS2']   = 0.
Fst7['TPitManS3']   = 0.
Fst7['TPitManE1']   = 0.
Fst7['TPitManE2']   = 0.
Fst7['TPitManE3']   = 0.
Fst7['BlPitch1']    = 0.
Fst7['BlPitch2']    = 0.
Fst7['BlPitch3']    = 0.
Fst7['BlPitchF1']   = 0.
Fst7['BlPitchF2']   = 0.
Fst7['BlPitchF3']   = 0.
Fst7['Gravity']     = 0.
Fst7['FlapDOF1']    = False
Fst7['FlapDOF2']    = False
Fst7['EdgeDOF']     = False
Fst7['TeetDOF']     = False
Fst7['DrTrDOF']     = False
Fst7['GenDOF']      = False
Fst7['YawDOF']      = False
Fst7['TwFADOF1']    = False
Fst7['TwFADOF2']    = False
Fst7['TwSSDOF1']    = False
Fst7['TwSSDOF2']    = False
Fst7['CompAero']    = False
Fst7['CompNoise']   = False
Fst7['OoPDefl']     = 0.
Fst7['IPDefl']      = 0.
Fst7['TeetDefl']    = 0.
Fst7['Azimuth']     = 0.
Fst7['RotSpeed']    = 0.
Fst7['NacYaw']      = 0.
Fst7['TTDspFA']     = 0.
Fst7['TTDspSS']     = 0.
Fst7['TipRad']      = 0.
Fst7['HubRad']      = 0.
Fst7['PSpnElN']     = 0.
Fst7['UndSling']    = 0
Fst7['HubCM']       = 0.
Fst7['OverHang']    = 0.
Fst7['NacCMxn']     = 0.
Fst7['NacCMy']      = 0.
Fst7['NacCMz']      = 0.
Fst7['TowerH']      = 0.
Fst7['Twr2Shft']    = 0.
Fst7['TwrRBHt']     = 0.
Fst7['ShftTilt']    = 0.
Fst7['Delta3']      = 0.
Fst7['PreCone(1)']    = 0.
Fst7['PreCone(2)']    = 0.
Fst7['PreCone(3)']    = 0.
Fst7['AzimB1Up']    = 0.
Fst7['YawBrMass']   = 0.
Fst7['NacMas']      = 0.
Fst7['HubMas']      = 0.
Fst7['TipMass(1)']    = 0.
Fst7['TipMass(2)']    = 0.
Fst7['TipMass(3)']    = 0.
Fst7['NacYIner']    = 0.
Fst7['GenIner']     = 0.
Fst7['HubIner']     = 0.
Fst7['GBoxEff']     = 0.
Fst7['GenEff']      = 0.
Fst7['GBRatio']     = 0.
Fst7['GBRevers']    = False
Fst7['HSSBrTqF']    = 0.
Fst7['HSSBrDT']     = 0.
Fst7['DynBrkFi']    = ""
Fst7['DTTorSpr']    = 0.
Fst7['DTTorDmp']    = 0.
Fst7['SIG_SlPc']    = 0.
Fst7['SIG_SySp']    = 0.
Fst7['SIG_RtTq']    = 0.
Fst7['SIG_PORt']    = 0.
Fst7['TEC_Freq']    = 0.
Fst7['TEC_NPol']    = 0
Fst7['TEC_SRes']    = 0.
Fst7['TEC_RRes']    = 0.
Fst7['TEC_VLL']     = 0.
Fst7['TEC_SLR']     = 0.
Fst7['TEC_RLR']     = 0.
Fst7['TEC_MR']      = 0.
Fst7['PtfmModel']   = 0
Fst7['PtfmFile']    = ""
Fst7['TwrNodes']    = 0
Fst7['TwrFile']     = ""
Fst7['YawSpr']      = 0.0
Fst7['YawDamp']     = 0.0
Fst7['YawNeut']     = 0.0
Fst7['Furling']     = False
Fst7['FurlFile']    = ""
Fst7['TeetMod']     = 0
Fst7['TeetDmpP']    = 0.
Fst7['TeetDmp']     = 0.
Fst7['TeetCDmp']    = 0.
Fst7['TeetSStP']    = 0.
Fst7['TeetHStP']    = 0.
Fst7['TeetSSSp']    = 0.
Fst7['TeetHSSp']    = 0.
Fst7['TBDrConN']    = 0.
Fst7['TBDrConD']    = 0.
Fst7['TpBrDT']      = 0.
Fst7['BldFile1']    = ""
Fst7['BldFile2']    = ""
Fst7['BldFile3']    = ""
Fst7['ADFile']      = ""
Fst7['NoiseFile']   = ""
Fst7['ADAMSFile']   = ""
Fst7['LinFile']     = ""
Fst7['SumPrint']    = False
Fst7['OutFileFmt']  = 1
Fst7['TabDelim']    = False
Fst7['OutFmt']      = ""
Fst7['TStart']      = 0.0
Fst7['DecFact']     = 0.
Fst7['SttsTime']    = 0.
Fst7['NcIMUxn']     = 0.
Fst7['NcIMUyn']     = 0.
Fst7['NcIMUzn']     = 0.
Fst7['ShftGagL']    = 0.
Fst7['NTwGages']    = 0
Fst7['TwrGagNd']    = 0
Fst7['NBlGages']    = 0
Fst7['BldGagNd']    = 0

# ====== INITIALIZE FAST MODEL BY INITIALIZING ALL VARIABLE TREES ======

FstModel = {}

# Description
FstModel['description']       = ''            

FstModel['Fst']               = Fst
FstModel['ElastoDyn']         = ElastoDyn
FstModel['ElastoDynBlade']    = ElastoDynBlade
FstModel['ElastoDynTower']    = ElastoDynTower
FstModel['InflowWind']        = InflowWind
FstModel['AeroDyn14']         = AeroDyn14
FstModel['AeroDyn15']         = AeroDyn15
FstModel['AeroDynBlade']      = AeroDynBlade
FstModel['AeroDynTower']      = AeroDynTower
FstModel['AeroDynPolar']      = AeroDynPolar
FstModel['ServoDyn']          = ServoDyn
FstModel['DISCON_in']         = DISCON_in
FstModel['HydroDyn']          = HydroDyn
FstModel['MoorDyn']           = MoorDyn
FstModel['SubDyn']            = SubDyn
FstModel['MAP']               = MAP
FstModel['BeamDyn']           = BeamDyn
FstModel['BeamDynBlade']      = BeamDynBlade
FstModel['Fst7']              = Fst7
        
# List of Outputs (all input files -- FST, ED, SD)
# TODO: Update FstOutput for a few new outputs in FAST8
FstModel['outlist']           = FstOutput   #
FstModel['outlist7']          = Fst7Output





'''
Regression testing for ROSCO

Tests:
    restart
'''    

import os
import platform
import unittest
import numpy as np

# Python Modules
import os
import platform
from shutil import copyfile

# ROSCO toolbox modules
from ROSCO_toolbox.inputs.validation import load_rosco_yaml
from ROSCO_toolbox.ofTools.fast_io import output_processing 
from ROSCO_toolbox.ofTools.case_gen.CaseLibrary import set_channels
from ROSCO_toolbox.ofTools.case_gen.CaseGen_General import CaseGen_General
from ROSCO_toolbox.ofTools.fast_io.FAST_writer import InputReader_OpenFAST, InputWriter_OpenFAST
from ROSCO_toolbox.utilities import run_openfast


class RegressionTesting(unittest.TestCase):
    def test_restart(self):
        this_dir = os.path.dirname(os.path.abspath(__file__))
        rosco_dir = os.path.dirname(this_dir)
        test_out_dir = os.path.join(this_dir, 'test_out')

        # Load yaml file (Open Loop Case)
        parameter_filename = os.path.join(rosco_dir, 'Tune_Cases/IEA15MW.yaml')

        inps = load_rosco_yaml(parameter_filename)
        path_params = inps['path_params']
        turbine_params = inps['turbine_params']
        controller_params = inps['controller_params']

        # Set rosco_dll
        if platform.system() == 'Windows':
            rosco_dll = os.path.join(rosco_dir, 'ROSCO/build/libdiscon.dll')
        elif platform.system() == 'Darwin':
            rosco_dll = os.path.join(rosco_dir, 'ROSCO/build/libdiscon.dylib')
        else:
            rosco_dll = os.path.join(rosco_dir, 'ROSCO/build/libdiscon.so')

        case_inputs = {}
        case_inputs[('Fst', 'TMax')] = {'vals': [3.], 'group': 0}
        case_inputs[('ServoDyn', 'DLL_FileName')] = {'vals': [rosco_dll], 'group': 0}
        case_inputs[('Fst', 'ChkptTime')] = {'vals': [1.], 'group': 1}
        case_inputs[('Fst', 'OutFileFmt')] = {'vals': [2], 'group': 1}
        case_inputs[('Fst', 'DT')] = {'vals': [0.025], 'group': 0}
        case_inputs[('DISCON_in', 'LoggingLevel')] = {'vals': [2], 'group': 1}

        # Generate cases
        run_dir = os.path.join(test_out_dir, 'restart')
        if not os.path.exists(run_dir):
            os.makedirs(run_dir)

        case_list, case_name_list = CaseGen_General(case_inputs, dir_matrix=run_dir, namebase='iea15_restart')
        channels = set_channels()

        # write files
        reader = InputReader_OpenFAST()
        writer = InputWriter_OpenFAST()
        reader.FAST_InputFile = path_params['FAST_InputFile']
        reader.FAST_directory = os.path.realpath(os.path.join( rosco_dir, 'Tune_Cases', path_params['FAST_directory']))
        reader.execute()
        writer.fst_vt = reader.fst_vt
        writer.FAST_runDirectory = test_out_dir
        for case, case_name in zip(case_list, case_name_list):
            writer.FAST_namingOut = case_name
            writer.update(fst_update=case)
            writer.update_outlist(channels)
            writer.execute()

        # Run first case
        fastcall = 'openfast'
        run_openfast(
            test_out_dir,
            fastcall=fastcall,
            fastfile=writer.FAST_InputFileOut,
            chdir=True
        )
        copyfile(os.path.join(test_out_dir, writer.FAST_InputFileOut[:-3]+'outb'),
                os.path.join(test_out_dir, 'orig.outb') )
        # copyfile(os.path.join(test_out_dir, writer.FAST_InputFileOut[:-3]+'RO.dbg'),
        #         os.path.join(test_out_dir, 'orig.dbg') )

        # run restart case
        run_openfast(
            test_out_dir,
            fastcall=fastcall,
            fastfile=writer.FAST_InputFileOut[:-3]+'40',
            chdir=True,
            restart=True
        )
        copyfile(os.path.join(test_out_dir, writer.FAST_InputFileOut[:-3]+'outb'),
            os.path.join(test_out_dir, 'restart.outb'))
        # copyfile(os.path.join(test_out_dir, writer.FAST_InputFileOut[:-3]+'RO.dbg'),
        #     os.path.join(test_out_dir, 'restart.dbg'))

        output_files = [os.path.join(test_out_dir, 'orig.outb'), 
                        os.path.join(test_out_dir, 'restart.outb')]
        op = output_processing.output_processing()
        fastout = op.load_fast_out(output_files, tmin=0)

        if False: # Plotting for debug
            import matplotlib.pyplot as plt
            cases = {}
            cases['Baseline'] = ['Wind1VelX', 'BldPitch1', 'GenTq', 'RotSpeed', 'NacYaw']
            fig, ax = op.plot_fast_out(cases=cases, showplot=False)
            plt.show()

        self.check_relative_error(fastout[1]['GenPwr'], fastout[0]['GenPwr'], 1e-3)

    def check_relative_error(self, meas, real, tol):
        '''check relative error'''
        error = np.linalg.norm(meas - real) / np.linalg.norm(real)
        self.assertTrue(error<=tol)

if __name__ == "__main__":
    unittest.main()



"""

Example script to run the DLCs in OpenFAST

"""

from ROSCO_toolbox.ofTools.case_gen.runFAST_pywrapper   import runFAST_pywrapper, runFAST_pywrapper_batch
from ROSCO_toolbox.ofTools.case_gen.CaseGen_IEC         import CaseGen_IEC
from ROSCO_toolbox.ofTools.case_gen.CaseGen_General     import CaseGen_General
from ROSCO_toolbox.ofTools.case_gen.CaseLibrary         import power_curve, set_channels, find_max_group, sweep_rated_torque, load_tuning_yaml, simp_step
from wisdem.commonse.mpi_tools              import MPI
import sys, os, platform
import numpy as np
from ROSCO_toolbox import utilities as ROSCO_utilities
from ROSCO_toolbox.inputs.validation import load_rosco_yaml

from ROSCO_toolbox import controller as ROSCO_controller
from ROSCO_toolbox import turbine as ROSCO_turbine

# Globals
this_dir        = os.path.dirname(os.path.abspath(__file__))
tune_case_dir   = os.path.realpath(os.path.join(this_dir,'../../../Tune_Cases'))


def run_FAST(tuning_yaml,wind_case_fcn,control_sweep_fcn,save_dir,n_cores=1):
    # set up run directory
    if control_sweep_fcn:
        sweep_name = control_sweep_fcn.__name__
    else:
        sweep_name = 'base'

    turbine_name = os.path.split(tuning_yaml)[-1].split('.')[0]
    run_dir = os.path.join(save_dir,turbine_name,wind_case_fcn.__name__,sweep_name)

    
    # Start with tuning yaml definition of controller
    if not os.path.isabs(tuning_yaml):
        tuning_yaml = os.path.join(tune_case_dir,tuning_yaml)


    # Load yaml file 
    inps = load_rosco_yaml(tuning_yaml)
    path_params         = inps['path_params']
    turbine_params      = inps['turbine_params']
    controller_params   = inps['controller_params']

    # Instantiate turbine, controller, and file processing classes
    turbine         = ROSCO_turbine.Turbine(turbine_params)
    controller      = ROSCO_controller.Controller(controller_params)

    # Load turbine data from OpenFAST and rotor performance text file
    cp_filename = os.path.join(tune_case_dir,path_params['FAST_directory'],path_params['rotor_performance_filename'])
    turbine.load_from_fast(path_params['FAST_InputFile'], \
        os.path.join(tune_case_dir,path_params['FAST_directory']), \
        dev_branch=True,rot_source='txt',\
        txt_filename=cp_filename)

    # tune base controller defined by the yaml
    controller.tune_controller(turbine)

    # Apply all discon variables as case inputs
    discon_vt = ROSCO_utilities.DISCON_dict(turbine, controller, txt_filename=cp_filename)
    control_base_case = {}
    for discon_input in discon_vt:
        control_base_case[('DISCON_in',discon_input)] = {'vals': [discon_vt[discon_input]], 'group': 0}

    # Set up wind case
    case_inputs = wind_case_fcn(run_dir)
    case_inputs.update(control_base_case)

    # Specify rosco controller dylib
    rosco_dll = '/Users/dzalkind/Tools/ROSCO/ROSCO/build/libdiscon.dylib' #'/Users/dzalkind/Tools/ROSCO_toolbox/ROSCO/build/libdiscon.dylib'

    if not rosco_dll: # use WEIS ROSCO
        run_dir1            = os.path.dirname( os.path.dirname( os.path.dirname( os.path.realpath(__file__) ) ) ) + os.sep
        if platform.system() == 'Windows':
            rosco_dll = os.path.join(run_dir1, 'local/lib/libdiscon.dll')
        elif platform.system() == 'Darwin':
            rosco_dll = os.path.join(run_dir1, 'local/lib/libdiscon.dylib')
        else:
            rosco_dll = os.path.join(run_dir1, 'local/lib/libdiscon.so')

    case_inputs[('ServoDyn','DLL_FileName')] = {'vals': [rosco_dll], 'group': 0}

    # Sweep control parameter
    if control_sweep_fcn:
        case_inputs_control = control_sweep_fcn(tuning_yaml,find_max_group(case_inputs)+1)
        sweep_name = control_sweep_fcn.__name__
        case_inputs.update(case_inputs_control)
    else:
        sweep_name = 'base'

    
        
    # Generate cases
    case_list, case_name_list = CaseGen_General(case_inputs, dir_matrix=run_dir, namebase=turbine_name)
    channels = set_channels()

    # Management of parallelization, leave in for now
    if MPI:
        from wisdem.commonse.mpi_tools import map_comm_heirarchical, subprocessor_loop, subprocessor_stop
        n_OF_runs = len(case_list)

        available_cores = MPI.COMM_WORLD.Get_size()
        n_parallel_OFruns = np.min([available_cores - 1, n_OF_runs])
        comm_map_down, comm_map_up, color_map = map_comm_heirarchical(1, n_parallel_OFruns)
        sys.stdout.flush()


    # Parallel file generation with MPI
    if MPI:
        comm = MPI.COMM_WORLD
        rank = comm.Get_rank()
    else:
        rank = 0
    if rank == 0:

        # Run FAST cases
        fastBatch                   = runFAST_pywrapper_batch(FAST_ver='OpenFAST',dev_branch = True)
        
        # Select Turbine Model
        model_dir                   = os.path.join(os.path.dirname( os.path.dirname( os.path.realpath(__file__) ) ), '01_aeroelasticse/OpenFAST_models')

        # FAST_directory (relative to Tune_Dir/)
        fastBatch.FAST_directory    = os.path.realpath(os.path.join(tune_case_dir,path_params['FAST_directory']))
        fastBatch.FAST_InputFile    = path_params['FAST_InputFile']        
        fastBatch.channels          = channels
        fastBatch.FAST_runDirectory = run_dir
        fastBatch.case_list         = case_list
        fastBatch.case_name_list    = case_name_list
        fastBatch.debug_level       = 2
        fastBatch.FAST_exe          = 'openfast'

        if MPI:
            fastBatch.run_mpi(comm_map_down)
        else:
            if n_cores == 1:
                fastBatch.run_serial()
            else:
                fastBatch.run_multi(cores=n_cores)

    if MPI:
        sys.stdout.flush()
        if rank in comm_map_up.keys():
            subprocessor_loop(comm_map_up)
        sys.stdout.flush()

    # Close signal to subprocessors
    if rank == 0 and MPI:
        subprocessor_stop(comm_map_down)
    sys.stdout.flush()
    

if __name__ == "__main__":

    # Simulation config
    sim_config = 6
    n_cores = 8

    if sim_config == 1:
        # FOCAL single wind speed testing
        tuning_yaml = '/Users/dzalkind/Tools/ROSCO/Tune_Cases/IEA15MW_FOCAL.yaml'
        wind_case   = simp_step
        sweep_mode  = None
        save_dir    = '/Users/dzalkind/Projects/FOCAL/torque_274'
    
    elif sim_config == 6:

        # FOCAL rated wind speed tuning
        tuning_yaml = '/Users/dzalkind/Tools/ROSCO/Tune_Cases/IEA15MW_FOCAL.yaml'
        wind_case   = power_curve
        sweep_mode  = sweep_rated_torque
        save_dir    = '/Users/dzalkind/Projects/FOCAL/drop_torque'

    else:
        raise Exception('This simulation configuration is not supported.')




    run_FAST(tuning_yaml,wind_case,sweep_mode,save_dir,n_cores=n_cores)

    # # Options: simp, pwr_curve
    # test_type   = 'pwr_curve'

    # save_dir_list    = [os.path.join(res_dir,tm,os.path.basename(dl).split('.')[0],test_type) \
    #     for tm, dl in zip(turbine_mods,discon_list)]

    # for tm, co, sd in zip(turbine_mods,discon_list,save_dir_list):
    #     run_Simp(tm,co,sd,n_cores=8)
    
    
    
    
